import os
import re
import select
import struct
import subprocess
import sys
import time
import typing as t
from pathlib import Path
import pytest
from functools import cached_property
import re
import subprocess
import platform
from django.test import TestCase

from django_typer.utils import detect_shell

from django_typer.management import get_command
from django_typer.management.commands.shellcompletion import Command as ShellCompletion
from django_typer.shells import DjangoTyperShellCompleter
from django_typer.utils import with_typehint
from ..utils import rich_installed, manage_py

default_shell = None

try:
    default_shell = detect_shell()[0]
except Exception:
    pass


def read_all_from_fd_with_timeout(fd, timeout=3):
    all_data = bytearray()
    start_time = time.time()

    while True:
        # Calculate remaining time
        remaining_time = timeout - (time.time() - start_time)
        if remaining_time <= 0:
            break  # Timeout reached

        # Wait until the file descriptor is ready or the timeout is reached
        rlist, _, _ = select.select([fd], [], [], remaining_time)
        if not rlist:
            break  # Timeout reached

        # Read available data
        data = os.read(fd, 1024)
        if not data:
            break  # End of file reached
        all_data.extend(data)

    return bytes(all_data).decode()


def scrub(output: str) -> str:
    """Scrub control code characters and ansi escape sequences for terminal colors from output"""
    return (
        re.sub(r"\x1B\[[0-?]*[ -/]*[@-~]", "", output, flags=re.IGNORECASE)
        .replace("\t", "")
        .replace("\x08", "")
    )


class _CompleteTestCase(with_typehint(TestCase)):
    shell: str
    manage_script: str
    launch_script: str

    interactive_opt: t.Optional[str] = None

    environment: t.List[str] = []

    tabs: str

    @cached_property
    def command(self) -> ShellCompletion:
        cmd = get_command("shellcompletion", ShellCompletion)
        cmd.init(shell=self.shell)
        return cmd

    def get_completer(self, **kwargs) -> DjangoTyperShellCompleter:
        return self.command.shell_class(
            **{"prog_name": self.manage_script, "command": self.command, **kwargs}
        )

    def setUp(self):
        self.remove()
        super().setUp()

    def tearDown(self):
        self.remove()
        super().tearDown()

    def verify_install(self, script=None, directory: t.Optional[Path] = None):
        pass

    def verify_remove(self, script=None, directory: t.Optional[Path] = None):
        pass

    def install(
        self,
        script=None,
        force_color=False,
        no_color=None,
        fallback=None,
        no_shell=False,
        prompt=False,
    ):
        if not script:
            script = self.manage_script
        init_kwargs = {"force_color": force_color, "no_color": no_color}
        kwargs = {"prompt": prompt}
        if script:
            kwargs["manage_script"] = script
        if self.shell and not no_shell:
            init_kwargs["shell"] = self.shell
        if fallback:
            kwargs["fallback"] = fallback
        self.command.init(**init_kwargs)
        self.command.install(**kwargs)
        self.verify_install(script=script)

    def remove(self, script=None):
        if not script:
            script = self.manage_script
        kwargs = {}
        if script:
            kwargs["manage_script"] = script
        if self.shell:
            self.command.init(shell=self.shell)
        self.command.uninstall(**kwargs)
        self.get_completions("ping")  # just to reinit shell
        self.verify_remove(script=script)

    if platform.system() == "Windows":

        def get_completions(self, *cmds: str, scrub_output=True, position=0) -> str:
            import winpty

            assert self.shell

            pty = winpty.PTY(256, 512)

            def read_all() -> str:
                output = ""
                while data := pty.read():
                    output += data
                    time.sleep(0.1)
                return output

            # Start the subprocess
            pty.spawn(
                self.shell, *([self.interactive_opt] if self.interactive_opt else [])
            )

            # Wait for the shell to start and get to the prompt
            time.sleep(3)
            read_all()

            for line in self.environment:
                pty.write(f"{line}{os.linesep}")

            time.sleep(2)
            output = read_all() + read_all()

            pty.write(" ".join(cmds))
            pty.write(position * ("\x1b[C" if position > 0 else "\x1b[D"))
            time.sleep(0.1)
            pty.write(self.tabs)

            time.sleep(2)
            completion = read_all() + read_all()

            return scrub(completion) if scrub_output else completion

    else:

        def get_completions(self, *cmds: str, scrub_output=True, position=0) -> str:
            import fcntl
            import termios
            import pty

            def read(fd):
                """Function to read from a file descriptor."""
                return os.read(fd, 1024 * 1024).decode()

            # Create a pseudo-terminal
            master_fd, slave_fd = pty.openpty()

            # Define window size - width and height
            os.set_blocking(slave_fd, False)
            win_size = struct.pack("HHHH", 24, 80, 0, 0)  # 24 rows, 80 columns
            fcntl.ioctl(slave_fd, termios.TIOCSWINSZ, win_size)

            # env = os.environ.copy()
            # env["TERM"] = "xterm-256color"

            # Spawn a new shell process
            shell = self.shell or detect_shell()[0]
            process = subprocess.Popen(
                [shell, *([self.interactive_opt] if self.interactive_opt else [])],
                stdin=slave_fd,
                stdout=slave_fd,
                stderr=slave_fd,
                text=True,
                # env=env,
                # preexec_fn=os.setsid,
            )
            # Wait for the shell to start and get to the prompt
            print(read(master_fd))

            for line in self.environment:
                os.write(master_fd, f"{line}{os.linesep}".encode())

            print(read(master_fd))
            # Send a command with a tab character for completion

            cmd = " ".join(cmds)
            os.write(master_fd, cmd.encode())

            # positioning does not seem to work :(
            if position < 0:
                os.write(master_fd, f"\033[{abs(position)}D".encode())
            elif position > 0:
                os.write(master_fd, f"\033[{abs(position)}C".encode())
            time.sleep(0.5)

            print(f'"{cmd}"')
            os.write(master_fd, self.tabs.encode())

            time.sleep(0.5)

            # Read the output
            output = read_all_from_fd_with_timeout(master_fd)

            # Clean up
            os.close(slave_fd)
            os.close(master_fd)
            process.terminate()
            process.wait()
            # remove bell character which can show up in some terminals where we hit tab
            if scrub_output:
                return scrub(output)
            return output

    def run_app_completion(self):
        completions = self.get_completions(self.launch_script, "completion", " ")
        self.assertIn("django_typer", completions)
        self.assertIn("admin", completions)
        self.assertIn("auth", completions)
        self.assertIn("contenttypes", completions)
        self.assertIn("messages", completions)
        self.assertIn("sessions", completions)
        self.assertIn("staticfiles", completions)

    def run_bad_command_completion(self):
        completions = self.get_completions(
            self.launch_script, "completion_does_not_exist", " "
        )
        self.assertTrue("Exception" not in completions)
        self.assertTrue("traceback" not in completions)

    def run_command_completion(self):
        completions = self.get_completions(self.launch_script, "complet")
        self.assertIn("completion", completions)
        completions = self.get_completions(self.launch_script, " ")
        self.assertIn("changepassword", completions)
        self.assertIn("check", completions)
        self.assertIn("dumpdata", completions)
        self.assertIn("completion", completions)
        self.assertIn("collectstatic", completions)

    def run_rich_option_completion(self, rich_output_expected: bool):
        completions = self.get_completions(
            self.launch_script, "completion", "--cmd", scrub_output=False
        )
        self.assertIn("--cmd", completions)
        self.assertIn("--cmd-first", completions)
        self.assertIn("--cmd-dup", completions)
        if not rich_installed:
            if self.shell not in ["powershell", "pwsh"]:
                self.assertIn("[bold]", completions)
                self.assertIn("[/bold]", completions)
                self.assertIn("[reverse]", completions)
                self.assertIn("[/reverse]", completions)
                self.assertIn("[underline]", completions)
                self.assertIn("[/underline]", completions)
                self.assertIn("[yellow]", completions)
                self.assertIn("[/yellow]", completions)
            else:
                self.assertTrue(
                    "[bold]" in completions
                    or "[/bold]" in completions
                    or "[reverse]" in completions
                    or "[/reverse]" in completions
                    or "[underline]" in completions
                    or "[/underline]" in completions
                    or "[yellow]" in completions
                    or "[/yellow]" in completions
                )
        elif rich_output_expected:
            # \x1b[0m and \x1b[m are the same
            if self.shell not in ["powershell", "pwsh"]:
                # exempt powershell from this because it filters the codes anyway
                self.assertIn("\x1b[7mcommands\x1b[", completions)
                self.assertIn("\x1b[4;33mcommands\x1b[", completions)
                self.assertIn("\x1b[1mname\x1b[", completions)
        else:
            self.assertNotIn("\x1b[7mcommands\x1b[", completions)
            self.assertNotIn("\x1b[4;33mcommands\x1b[", completions)
            self.assertNotIn("\x1b[1mimport path\x1b[", completions)
            self.assertNotIn("\x1b[1mname\x1b[", completions)

    def test_shell_complete(self):
        with self.assertRaises(AssertionError):
            self.run_app_completion()
        self.install()
        self.run_app_completion()
        self.run_bad_command_completion()
        self.run_command_completion()
        self.remove()
        with self.assertRaises(AssertionError):
            self.run_app_completion()
        self.install()

    def test_fallback(self):
        self.remove()
        self.install(fallback="tests.fallback.custom_fallback")
        completions = self.get_completions(self.launch_script, " ")
        self.assertIn("custom_fallback", completions)

    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_rich_output(self):
        self.install(force_color=True)
        self.run_rich_option_completion(rich_output_expected=True)

    @pytest.mark.rich
    @pytest.mark.skipif(not rich_installed, reason="Rich not installed")
    def test_no_rich_output(self):
        self.install(no_color=True)
        self.run_rich_option_completion(rich_output_expected=False)

    def test_settings_pass_through(self):
        # https://github.com/django-commons/django-typer/issues/68
        self.install()
        completions = self.get_completions(self.launch_script, "app_labels", " ")
        self.assertNotIn("django_typer", completions)
        completions = self.get_completions(
            self.launch_script,
            "app_labels",
            "--settings",
            "tests.settings.examples",
            " ",
        )
        self.assertIn("django_typer", completions)
        completions = self.get_completions(
            self.launch_script,
            "app_labels",
            "--settings=tests.settings.examples",
            " ",
        )
        self.assertIn("django_typer", completions)

    def test_pythonpath_pass_through(self):
        # https://github.com/django-commons/django-typer/issues/68
        self.install()
        completions = self.get_completions(
            self.launch_script, "python_path", "--options", " "
        )
        self.assertNotIn("working", completions)
        completions = self.get_completions(
            self.launch_script,
            "python_path",
            "--pythonpath",
            "tests/off_path",
            "--option",
            " ",
        )
        self.assertIn("working", completions)
        completions = self.get_completions(
            self.launch_script,
            "python_path",
            "--pythonpath=tests/off_path",
            "--option",
            " ",
        )
        self.assertIn("working", completions)

    def test_reentrant_install_uninstall(self):
        self.install()
        self.install()
        self.verify_install()

        completions = self.get_completions(self.launch_script, "complet")
        self.assertIn("completion", completions)

        self.remove()
        self.remove()
        self.verify_remove()

    def test_path_completion(self):
        self.install()
        self.verify_install()
        completions = self.get_completions(
            self.launch_script, "completion", "--path", "./src/django_typer/co"
        )
        self.assertIn("completers", completions)
        self.assertIn("config.py", completions)
        completions = self.get_completions(
            self.launch_script, "completion", "--dir", "./src/django_typer/"
        )
        self.assertNotIn("utils.py", completions)
        self.assertNotIn("config.py", completions)
        self.assertIn("templates", completions)
        self.assertIn("management", completions)
        self.remove()
        self.verify_remove()

    # todo - cursor positioning not working
    # def test_cursor_position(self):
    #     self.install()
    #     self.verify_install()
    #     cmd = [self.launch_script, "shellcompletion", "--set ", "install"]
    #     completions = self.get_completions(*cmd, position=-9)
    #     self.assertIn("--settings", completions)
    #     self.remove()
    #     self.verify_remove()


class _ScriptCompleteTestCase(_CompleteTestCase):
    manage_script: str = "manage.py"
    launch_script: str = "./manage.py"


class _InstalledScriptCompleteTestCase(_CompleteTestCase):
    """
    These shell completes use an installed script available on the path
    instead of a script directly invoked by path. The difference may
    seem trivial - but it is not given how most shells determine if completion
    logic should be invoked for a given command.
    """

    MANAGE_SCRIPT_TMPL = Path(__file__).parent / "django_manage.py"
    manage_script = "django-admin"
    launch_script = "django-admin"

    @classmethod
    def install_script(cls, script=None):
        if not script:
            script = cls.manage_script
        lines = []
        with open(cls.MANAGE_SCRIPT_TMPL, "r") as f:
            for line in f.readlines():
                if line.startswith("#!{{shebang}}"):
                    line = f"#!{sys.executable}\n"
                lines.append(line)
        exe = Path(sys.executable).parent / script
        with open(exe, "w") as f:
            for line in lines:
                f.write(line)

        # make the script executable
        os.chmod(exe, os.stat(exe).st_mode | 0o111)

        if platform.system() == "Windows":
            with open(exe.with_suffix(".cmd"), "w") as f:
                f.write(f'@echo off{os.linesep}"{sys.executable}" "%~dp0{exe.name}" %*')
            os.chmod(exe, os.stat(exe.with_suffix(".cmd")).st_mode | 0o111)

    @classmethod
    def remove_script(cls, script=None):
        if not script:
            script = cls.manage_script
        exe = Path(sys.executable).parent / script
        exe.unlink(missing_ok=True)
        exe.with_suffix(".cmd").unlink(missing_ok=True)

    def test_multi_install(self):
        parts = self.manage_script.split(".")
        manage2 = ".".join([parts[0] + "2", *parts[1:]])
        try:
            self.install_script(script=manage2)
            self.install()
            self.verify_install()
            self.install(script=manage2)
            self.verify_install(script=manage2)

            completions = self.get_completions(self.manage_script, "complet")
            self.assertIn("completion", completions)

            completions = self.get_completions(manage2, "complet")
            self.assertIn("completion", completions)

            self.remove()
            self.verify_remove()
            self.remove(script=manage2)
            self.verify_remove(script=manage2)
        finally:
            self.remove_script(script=manage2)

    def test_prompt_install(self, env={}, directory: t.Optional[Path] = None):
        import pexpect

        env = {
            **dict(os.environ),
            "DJANGO_SETTINGS_MODULE": "tests.settings.completion",
            "DJANGO_COLORS": "nocolor",
            **env,
        }

        rex = re.compile
        expected = [
            rex(
                r"Append\s+the\s+above\s+contents\s+to\s+(?P<file>.*)\?", re.DOTALL
            ),  # 0
            rex(
                r"Create\s+(?P<file>.*)\s+with\s+the\s+above\s+contents\?",
                re.DOTALL,
            ),  # 1
            rex(r"Aborted\s+shell\s+completion\s+installation."),  # 2
            rex(rf"Installed\s+autocompletion\s+for\s+{self.shell}"),  # 3
        ]

        install_command = [
            "shellcompletion",
            "--no-color",
            "--shell",
            self.shell,
            "install",
        ]
        self.remove()
        self.verify_remove(directory=directory)

        if platform.system() != "Windows":
            install = pexpect.spawn(self.manage_script, install_command, env=env)
            install.setwinsize(24, 800)
        else:
            from pexpect.popen_spawn import PopenSpawn

            install = PopenSpawn(
                " ".join([self.manage_script, *install_command]),
                env=env,
                encoding="utf-8",
            )

        def wait_for_output(child) -> t.Tuple[int, t.Optional[str]]:
            index = child.expect(expected)
            if index in [0, 1]:
                return index, child.match.group("file")
            return index, None

        # test an abort
        idx, _ = wait_for_output(install)
        self.assertLess(idx, 2)
        install.sendline("N")

        while True:
            idx, _ = wait_for_output(install)
            if idx < 2:
                install.sendline("N")
            else:
                self.assertEqual(idx, 2)
                break

        self.verify_remove(directory=directory)

        # test an install
        if platform.system() != "Windows":
            install = pexpect.spawn(self.manage_script, install_command, env=env)
            install.setwinsize(24, 800)
        else:
            from pexpect.popen_spawn import PopenSpawn

            install = PopenSpawn(
                " ".join([self.manage_script, *install_command]),
                env=env,
                encoding="utf-8",
            )

        while True:
            idx, _ = wait_for_output(install)
            if idx < 2:
                install.sendline("Y")
            else:
                self.assertEqual(idx, 3)
                break

        self.verify_install(directory=directory)

    # TODO
    # else:

    #     def test_prompt_install(self, env={}, directory: t.Optional[Path] = None):
    #         env = {
    #             **dict(os.environ),
    #             "DJANGO_SETTINGS_MODULE": "tests.settings.completion",
    #             "DJANGO_COLORS": "nocolor",
    #             **env,
    #         }

    #         rex = re.compile
    #         expected_patterns = [
    #             rex(r"Append the above contents to (?P<file>.*)\?"),  # 0
    #             rex(r"Create (?P<file>.*) with the above contents\?"),  # 1
    #             rex(r"Aborted shell completion installation."),  # 2
    #             rex(rf"Installed autocompletion for {self.shell}"),  # 3
    #         ]

    #         install_command = [
    #             self.manage_script,
    #             "shellcompletion",
    #             "--no-color",
    #             "--shell",
    #             self.shell,
    #             "install",
    #         ]
    #         self.remove()
    #         self.verify_remove(directory=directory)

    #         def run_with_response(responses: t.List[str]):
    #             process = subprocess.Popen(
    #                 install_command,
    #                 env=env,
    #                 cwd=directory,
    #                 stdin=subprocess.PIPE,
    #                 stdout=subprocess.PIPE,
    #                 stderr=subprocess.STDOUT,
    #                 text=True,
    #             )

    #             output = ""
    #             for response in responses:
    #                 while True:
    #                     line = process.stdout.readline()
    #                     if not line:
    #                         break
    #                     output += line

    #                     matched_index, matched_file = match_output(line)
    #                     if matched_index is not None:
    #                         process.stdin.write(response + "\n")
    #                         process.stdin.flush()
    #                         break

    #             process.wait()
    #             return output

    #         def match_output(line: str) -> t.Tuple[t.Optional[int], t.Optional[str]]:
    #             for i, pattern in enumerate(expected_patterns):
    #                 match = pattern.search(line)
    #                 if match:
    #                     return i, match.groupdict().get("file")
    #             return None, None

    #         # Test abort sequence
    #         abort_output = run_with_response(["N", "N"])
    #         self.assertIn("Aborted shell completion installation.", abort_output)
    #         self.verify_remove(directory=directory)

    #         # Test install sequence
    #         install_output = run_with_response(["Y", "Y"])
    #         self.assertIn(f"Installed autocompletion for {self.shell}", install_output)
    #         self.verify_install(directory=directory)



from tests.apps.test_app.management.commands.adapted2 import (
    Command as Adapted2,
)


@Adapted2.command()
def top1(self):
    """Extend with top level command."""
    assert self.__class__ is Adapted2
    return f"adapter2::adapted2({self.verbosity})::top1()"


@Adapted2.grp1.command()
def grp1_adpt1(self, int_arg1: int, int_arg1_2: int):
    """Extend grp1 with command."""
    assert self.__class__ is Adapted2
    return f"adapter2::adapted2({self.verbosity})::grp1({self.argg3})::grp1_adpt1({int_arg1}, {int_arg1_2})"


@Adapted2.grp2.group(invoke_without_command=True)
def sub_grp2(self, int_arg2: int, int_arg2_2: int):
    assert self.__class__ is Adapted2
    return f"adapter2::adapted2({self.verbosity})::grp2({self.flag1})::sub_grp2({int_arg2}, {int_arg2_2})"


try:

    @sub_grp2.group(invoke_without_command=True)
    def subsub_grp2():
        return "adapter2::adapted2()::grp2()::sub_grp2()::subsub_grp2()"

    assert False, (
        "Should not be able to do this, because we must always be very specific as to which command we are modifying"
    )
except AssertionError:
    pass


@Adapted2.sub_grp2.group(invoke_without_command=True)
def subsub_grp2():
    return "adapter2::adapted2()::grp2()::sub_grp2()::subsub_grp2()"


@Adapted2.subsubgroup.command()
def ssg_cmd(self, int_arg3: int, int_arg3_2: int):
    assert self.__class__ is Adapted2
    return f"adapter2::interference::grp1({self.argg3})::subgroup({self.arg5_0})::subsubgroup({self.subsubgroup_called})::ssg_cmd({int_arg3}, {int_arg3_2})"



from django_typer.management import Typer, TyperCommand
from django_typer.types import Verbosity

Command: TyperCommand

app = Typer()


_verbosity = None
_flag = None


@app.callback()
def init(verbosity: Verbosity = 0):
    global _verbosity
    _verbosity = verbosity


@app.command()
def main(name: str):
    return {"verbosity": _verbosity, "name": name}


grp2 = Typer()


@grp2.callback()
def init_grp1(flag: bool = False):
    global _flag
    _flag = flag


@grp2.command()
def cmd2(fraction: float):
    return {"verbosity": _verbosity, "flag": _flag, "fraction": fraction}


#  this works
app.add_typer(grp2, name="grp1")


@grp2.command()
def cmd1(count: int):
    return {"verbosity": _verbosity, "flag": _flag, "count": count}


def run_subgrp(msg: str):
    return {"verbosity": _verbosity, "flag": _flag, "msg": msg}


subgrp = Typer(name="subgrp", invoke_without_command=True, callback=run_subgrp)

grp2.add_typer(subgrp)



import os
import typing as t
from functools import cached_property
from pathlib import Path

from click.shell_completion import CompletionItem

from . import DjangoTyperShellCompleter


class ZshComplete(DjangoTyperShellCompleter):
    """
    This completer class supports Zsh_. Completion scripts are installed to the
    ``.zfunc`` directory in the user's home directory. Style and completion
    initialization instructions are added to the user's ``.zshrc`` file if needed.

    Returned suggestions are delimited by newlines. Each suggestion is on three lines:

    * The first line is the type of the completion.
    * The second line is the value of the completion.
    * The third line is the help text for the completion.

    See also:
    https://github.com/zsh-users/zsh-completions/blob/master/zsh-completions-howto.org
    """

    name = "zsh"
    """
    shell executable.
    """

    template = "shell_complete/zsh.sh"
    """
    The template used to render the zsh completion script.
    """

    supports_scripts = True
    """
    The zsh completer supports script invocations.
    """

    color = False
    """
    Zsh_ does support ansi control codes in completion suggestions, but we disable them
    by default.
    """

    def get_user_profile(self) -> Path:
        """
        Get the user's .zshrc file. It is located in the user's home directory by
        default unless the ``ZDOTDIR`` environment variable is set.
        """
        return Path(os.environ.get("ZDOTDIR", Path.home())) / ".zshrc"

    @cached_property
    def install_dir(self) -> Path:
        """
        The directory where completer scripts will be installed.
        """
        install_dir = self.get_user_profile().parent / ".zfunc"
        install_dir.mkdir(parents=True, exist_ok=True)
        return install_dir

    def format_completion(self, item: CompletionItem) -> str:
        hlp = self.process_rich_text(item.help.replace("\n", " ")) if item.help else "_"
        return f"{item.type}\n{self.process_rich_text(item.value)}\n{hlp}"

    def install(self, prompt: bool = True) -> t.List[Path]:
        assert self.prog_name
        zshrc = self.get_user_profile()
        zshrc_source = ""
        start_line = 0
        if zshrc.is_file():
            zshrc_source = zshrc.read_text()
            start_line = zshrc_source.count("\n") + 2
        additions = ""
        if "fpath" not in zshrc_source:
            additions += f"if type brew &>/dev/null; then{os.linesep}"
            additions += (
                f"\tfpath=(~/.zfunc $(brew --prefix)/share/zsh-completions $fpath)"
                f"{os.linesep}"
            )
            additions += f"else{os.linesep}"
            additions += f"\tfpath=(~/.zfunc $fpath){os.linesep}"
            additions += f"fi{os.linesep}{os.linesep}"
        if "compinit" not in zshrc_source:
            additions += f"autoload -Uz compinit{os.linesep}"
            additions += f"compinit{os.linesep}"

        style = (
            f"zstyle ':completion:*:*:{self.prog_name}:*' menu select"
            if self.is_installed
            else "zstyle ':completion:*' menu select"
        )
        if style not in zshrc_source:
            additions += f"{style}{os.linesep}"

        edited = []
        script = self.install_dir / f"_{self.prog_name}"
        if additions:
            if self.prompt(
                prompt=prompt,
                source=additions,
                file=zshrc,
                start_line=start_line,
            ):
                Path.home().mkdir(parents=True, exist_ok=True)
                with open(zshrc, "a") as zrc_file:
                    zrc_file.write(f"{os.linesep}{additions}")
                edited.append(zshrc)
        source = self.source()
        if self.prompt(prompt=prompt, source=source, file=script, start_line=0):
            script.write_text(self.source())
            edited.append(script)
        return edited

    def uninstall(self):
        script = self.install_dir / f"_{self.prog_name}"
        if script.is_file():
            script.unlink()

        zshrc = self.get_user_profile()
        if zshrc.is_file():
            zshrc_source = zshrc.read_text()
            zshrc.write_text(
                zshrc_source.replace(
                    f"zstyle ':completion:*:*:{self.prog_name}:*' menu select\n", ""
                )
            )



import os
import re
from io import StringIO

import pytest
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import manage_py, run_command


class TestHelpPrecedence(TestCase):
    def test_help_precedence1(self):
        buffer = StringIO()
        cmd = get_command("help_precedence1", no_color=True, stdout=buffer)
        cmd.print_help("./manage.py", "help_precedence1")
        self.assertTrue(
            re.search(
                r"help_precedence1\s+Test minimal TyperCommand subclass - command method",
                buffer.getvalue(),
            )
        )
        self.assertIn(
            "Test minimal TyperCommand subclass - typer param", buffer.getvalue()
        )

    def test_help_precedence2(self):
        buffer = StringIO()
        cmd = get_command("help_precedence2", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence2")
        self.assertIn(
            "Test minimal TyperCommand subclass - class member", buffer.getvalue()
        )
        self.assertTrue(
            re.search(
                r"help_precedence2\s+Test minimal TyperCommand subclass - command method",
                buffer.getvalue(),
            )
        )

    def test_help_precedence3(self):
        buffer = StringIO()
        cmd = get_command("help_precedence3", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence3")
        self.assertTrue(
            re.search(
                r"help_precedence3\s+Test minimal TyperCommand subclass - command method",
                buffer.getvalue(),
            )
        )
        self.assertIn(
            "Test minimal TyperCommand subclass - callback method", buffer.getvalue()
        )

    def test_help_precedence4(self):
        buffer = StringIO()
        cmd = get_command("help_precedence4", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence4")
        self.assertIn(
            "Test minimal TyperCommand subclass - callback docstring", buffer.getvalue()
        )
        self.assertTrue(
            re.search(
                r"help_precedence4\s+Test minimal TyperCommand subclass - command method",
                buffer.getvalue(),
            )
        )

    def test_help_precedence5(self):
        buffer = StringIO()
        cmd = get_command("help_precedence5", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence5")
        self.assertIn(
            "Test minimal TyperCommand subclass - command method", buffer.getvalue()
        )

    def test_help_precedence6(self):
        buffer = StringIO()
        cmd = get_command("help_precedence6", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence6")
        self.assertIn(
            "Test minimal TyperCommand subclass - docstring", buffer.getvalue()
        )

    def test_help_precedence7(self):
        buffer = StringIO()
        cmd = get_command("help_precedence7", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence7")
        self.assertIn(
            "Test minimal TyperCommand subclass - class member", buffer.getvalue()
        )

    def test_help_precedence8(self):
        buffer = StringIO()
        cmd = get_command("help_precedence8", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence8")
        self.assertIn(
            "Test minimal TyperCommand subclass - typer param", buffer.getvalue()
        )

    def test_help_precedence9(self):
        buffer = StringIO()
        cmd = get_command("help_precedence9", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence9")
        self.assertIn("Class docstring.", buffer.getvalue())

    def test_help_precedence10(self):
        buffer = StringIO()
        cmd = get_command("help_precedence10", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence10")
        self.assertIn("Class docstring.", buffer.getvalue())

    def test_help_precedence11(self):
        buffer = StringIO()
        cmd = get_command("help_precedence11", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence11")
        self.assertIn("This docstring overrides base docstring.", buffer.getvalue())

    def test_help_precedence12(self):
        buffer = StringIO()
        cmd = get_command("help_precedence12", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence12")
        self.assertIn("Class docstring.", buffer.getvalue())

    def test_help_precedence13(self):
        buffer = StringIO()
        cmd = get_command("help_precedence13", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence13")
        self.assertIn(
            "Test minimal TyperCommand subclass - typer param", buffer.getvalue()
        )
        self.assertNotIn("Override higher precedence inherit.", buffer.getvalue())

    def test_help_precedence14(self):
        buffer = StringIO()
        cmd = get_command("help_precedence14", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence14")
        self.assertIn(
            "Test minimal TyperCommand subclass - typer param", buffer.getvalue()
        )
        self.assertNotIn("Override higher precedence inherit.", buffer.getvalue())

    def test_help_precedence15(self):
        buffer = StringIO()
        cmd = get_command("help_precedence15", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence15")
        self.assertNotIn(
            "Test minimal TyperCommand subclass - typer param", buffer.getvalue()
        )
        self.assertIn("Now class docstring is used!", buffer.getvalue())

    def test_help_precedence16(self):
        buffer = StringIO()
        cmd = get_command("help_precedence16", stdout=buffer, no_color=True)
        cmd.print_help("./manage.py", "help_precedence16")
        self.assertNotIn(
            "Test minimal TyperCommand subclass - typer param", buffer.getvalue()
        )
        self.assertIn("Now class docstring is used!", buffer.getvalue())

    @pytest.mark.skip()
    def test_help_from_other_dir(self):
        """
        This test is for coverage that the help ouput resolves the correct script when
        called from a non-relative path.

        TODO - running from another directory screws up the code coverage b/c relative
        paths. Fix this.
        """
        cwd = os.getcwd()
        try:
            os.chdir("./django_typer/tests")
            stdout, _, retcode = run_command(
                "help_precedence9", "--no-color", "--help", chdir=False
            )
            self.assertEqual(retcode, 0)
            self.assertIn("Class docstring.", stdout)
            self.assertIn(f"Usage: {manage_py}", stdout)
        finally:
            os.chdir(cwd)


class TestNativeHelpPrecedence(TestCase):
    def test_help_precedence_native1(self):
        buffer = StringIO()
        cmd = get_command("help_precedence_native1", no_color=True, stdout=buffer)
        cmd.print_help("./manage.py", "help_precedence_native1")
        self.assertIn("1: Command help", buffer.getvalue())

    def test_help_precedence_native2(self):
        buffer = StringIO()
        cmd = get_command("help_precedence_native2", no_color=True, stdout=buffer)
        cmd.print_help("./manage.py", "help_precedence_native2")
        self.assertIn("2: App help", buffer.getvalue())

    def test_help_precedence_native3(self):
        buffer = StringIO()
        cmd = get_command("help_precedence_native3", no_color=True, stdout=buffer)
        cmd.print_help("./manage.py", "help_precedence_native3")
        self.assertIn("3: Docstring help", buffer.getvalue())



from django_typer.utils import register_command_plugins
from tests.apps.examples.plugins.backup.management.commands.backup import Command as Backup
from tests.apps.examples.plugins.media2.management import plugins as media2_plugins
from tests.apps.examples.plugins.my_app.management import plugins as my_app_plugins

register_command_plugins(media2_plugins)
register_command_plugins(my_app_plugins)

app = Backup().typer_app

# todo necessary?


r"""
::

      ██████╗      ██╗ █████╗ ███╗   ██╗ ██████╗  ██████╗
      ██╔══██╗     ██║██╔══██╗████╗  ██║██╔════╝ ██╔═══██╗
      ██║  ██║     ██║███████║██╔██╗ ██║██║  ███╗██║   ██║
      ██║  ██║██   ██║██╔══██║██║╚██╗██║██║   ██║██║   ██║
      ██████╔╝╚█████╔╝██║  ██║██║ ╚████║╚██████╔╝╚██████╔╝
      ╚═════╝  ╚════╝ ╚═╝  ╚═╝╚═╝  ╚═══╝ ╚═════╝  ╚═════╝

         ████████╗██╗   ██╗██████╗ ███████╗██████╗
         ╚══██╔══╝╚██╗ ██╔╝██╔══██╗██╔════╝██╔══██╗
            ██║    ╚████╔╝ ██████╔╝█████╗  ██████╔╝
            ██║     ╚██╔╝  ██╔═══╝ ██╔══╝  ██╔══██╗
            ██║      ██║   ██║     ███████╗██║  ██║
            ╚═╝      ╚═╝   ╚═╝     ╚══════╝╚═╝  ╚═╝


django-typer_ provides an extension class,
:class:`~django_typer.management.TyperCommand`, to the
:class:`~django.core.management.BaseCommand` class that melds the Typer_ and
:doc:`click <click:index>` infrastructure with the Django_ infrastructure. The result is
all the ease of specifying commands, groups and options and arguments using Typer_ and
:doc:`click <click:index>` in a way that feels like and is interface compatible with
Django_'s :class:`~django.core.management.BaseCommand` This should enable a smooth
transition for existing Django_ commands and an intuitive feel for implementing new
commands.

django-typer_ also supports shell completion for bash_, zsh_, fish_ and powershell_ and
extends that support to native Django_ management commands as well.


The goal of django-typer_ is to provide full Typer_ style functionality while
maintaining compatibility with the Django management command system. This means that the
:class:`~django.core.management.BaseCommand` interface is preserved and the Typer_
interface is added on top of it. This means that this code base is more robust to
changes in the Django management command system - because most of the base class
functionality is preserved but many Typer_ and :doc:`click <click:index>` internals are
used directly to achieve this. We rely on robust CI to catch breaking changes upstream
and keep a tight version lock on Typer.
"""

VERSION = (3, 6, 3)

__title__ = "Django Typer"
__version__ = ".".join(str(i) for i in VERSION)
__author__ = "Brian Kohan"
__license__ = "MIT"
__copyright__ = "Copyright 2023-2026 Brian Kohan"



import typing as t

from django.utils.translation import gettext_lazy as _
from typer import Argument

from django_typer.management import TyperCommand, command


class Command(TyperCommand):
    """
    A command that defines subcommands.
    """

    @command()
    def create(
        self,
        name: t.Annotated[
            str, Argument(help=_("The name of the object to create."))
        ],
    ):
        """
        Create an object.
        """

    @command()
    def delete(
        self,
        id: t.Annotated[
            int, Argument(help=_("The id of the object to delete."))
        ],
    ):
        """
        Delete an object.
        """



import json


from django_typer.management import TyperCommand


class Command(TyperCommand):
    """
    Class docstring.
    """

    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        assert self.__class__ is Command
        opts = {"arg1": arg1, "arg2": arg2, "arg3": arg3, "arg4": arg4}
        return json.dumps(opts)



import os
import typing as t
from functools import cached_property
from pathlib import Path

from click.shell_completion import CompletionItem

from . import DjangoTyperShellCompleter


class FishComplete(DjangoTyperShellCompleter):
    """
    This completer class supports the fish_ shell. Completion scripts are installed in
    the ``~/.config/fish/completions`` directory.

    Returned suggestions are formatted as ``type,value[\thelp]``. Each suggestion is on
    one line and if no help is provided, the help text including the tab delimiter is
    omitted.
    """

    name = "fish"
    """
    shell executable.
    """

    template = "shell_complete/fish.fish"
    """
    The template used to render the fish completion script.
    """

    supports_scripts = False
    """
    Fish does not support script invocations.
    """

    color = False
    """
    Fish does not support ansi control codes.
    """

    def get_user_profile(self) -> Path:
        """
        Get the user's fish config file. It is located in the user's home directory by
        default unless the ``XDG_CONFIG_HOME`` environment variable is set.
        """
        return (
            Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config"))
            / "fish/config.fish"
        )

    @cached_property
    def install_dir(self) -> Path:
        install_dir = self.get_user_profile().parent / "completions"
        install_dir.mkdir(parents=True, exist_ok=True)
        return install_dir

    def format_completion(self, item: CompletionItem) -> str:
        if item.help:
            return (
                f"{item.type},{self.process_rich_text(item.value)}\t"
                f"{self.process_rich_text(item.help)}"
            )
        return f"{item.type},{self.process_rich_text(item.value)}"

    def install(self, prompt: bool = True) -> t.List[Path]:
        assert self.prog_name
        script = self.install_dir / f"{self.prog_name}.fish"
        source = self.source()
        if self.prompt(prompt=prompt, source=source, file=script):
            script.write_text(source)
            return [script]
        return []

    def uninstall(self):
        assert self.prog_name
        script = self.install_dir / f"{self.prog_name}.fish"
        if script.is_file():
            script.unlink()



import json

from django.utils.translation import gettext_lazy as _

from django_typer.management import TyperCommand, command, initialize


class Command(TyperCommand, help=_("Test minimal TyperCommand subclass - typer param")):
    """
    Class docstring.
    """

    help = _("Test minimal TyperCommand subclass - class member")

    @initialize(help=_("Test minimal TyperCommand subclass - callback method"))
    def init(self, verbosity: int = 1):
        """
        Test minimal TyperCommand subclass - callback docstring
        """
        assert self.__class__ is Command
        self.verbosity = verbosity

    @command(help=_("Test minimal TyperCommand subclass - command method"))
    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        """
        Test minimal TyperCommand subclass - docstring
        """
        assert self.__class__ is Command
        opts = {"arg1": arg1, "arg2": arg2, "arg3": arg3, "arg4": arg4}
        return json.dumps(opts)



import pytest
from django.test import TestCase, override_settings

from django_typer.management import get_command
from tests.utils import rich_installed, run_command, similarity

basic_help = """
 Usage: ./manage.py basic [OPTIONS] ARG1 ARG2                                   
                                                                                
 A basic command that uses Typer                                                
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    arg1      TEXT  [required]                                              │
│ *    arg2      TEXT  [required]                                              │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --arg3        FLOAT    [default: 0.5]                                        │
│ --arg4        INTEGER  [default: 1]                                          │
│ --help                 Show this message and exit.                           │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --version                  Show program's version number and exit.           │
│ --settings           TEXT  The Python path to a settings module, e.g.        │
│                            "myproject.settings.main". If this isn't          │
│                            provided, the DJANGO_SETTINGS_MODULE environment  │
│                            variable will be used.                            │
│ --pythonpath         PATH  A directory to add to the Python path, e.g.       │
│                            "/home/djangoprojects/myproject".                 │
│ --traceback                Raise on CommandError exceptions                  │
│ --show-locals              Print local variables in tracebacks.              │
│ --no-color                 Don't colorize the command output.                │
│ --force-color              Force colorization of the command output.         │
│ --skip-checks              Skip system checks.                               │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

basic_help_no_rich = """
Usage: ./manage.py basic [OPTIONS] ARG1 ARG2

  A basic command that uses Typer

Arguments:
  ARG1  [required]
  ARG2  [required]

Options:
  --arg3 FLOAT       [default: 0.5]
  --arg4 INTEGER     [default: 1]
  --version          Show program's version number and exit.
  --settings TEXT    The Python path to a settings module, e.g.
                     "myproject.settings.main". If this isn't provided, the
                     DJANGO_SETTINGS_MODULE environment variable will be used.
  --pythonpath PATH  A directory to add to the Python path, e.g.
                     "/home/djangoprojects/myproject".
  --traceback        Raise on CommandError exceptions
  --no-color         Don't colorize the command output.
  --force-color      Force colorization of the command output.
  --skip-checks      Skip system checks.
  --help             Show this message and exit.
"""

multi_help_no_rich = """
Usage: ./manage.py multi [OPTIONS] COMMAND [ARGS]...
  
  A command that defines subcommands.

Options:
  --version          Show program's version number and exit.
  --settings TEXT    The Python path to a settings module, e.g.
                     "myproject.settings.main". If this isn't provided, the
                     DJANGO_SETTINGS_MODULE environment variable will be used.
  --pythonpath PATH  A directory to add to the Python path, e.g.
                     "/home/djangoprojects/myproject".
  --traceback        Raise on CommandError exceptions
  --no-color         Don't colorize the command output.
  --force-color      Force colorization of the command output.
  --skip-checks      Skip system checks.
  --help             Show this message and exit.

Commands:
  create  Create an object.
  delete  Delete an object.
"""

multi_create_help_no_rich = """
Usage: ./manage.py multi create [OPTIONS] NAME

  Create an object.

Arguments:
  NAME  The name of the object to create.  [required]

Options:
  --help  Show this message and exit.
"""

multi_delete_help_no_rich = """
Usage: ./manage.py multi delete [OPTIONS] ID

  Delete an object.

Arguments:
  ID  The id of the object to delete.  [required]

Options:
  --help  Show this message and exit.
"""

multi_help = """
 Usage: ./manage.py multi [OPTIONS] COMMAND [ARGS]...                           
 
  A command that defines subcommands.

╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --version                  Show program's version number and exit.           │
│ --settings           TEXT  The Python path to a settings module, e.g.        │
│                            "myproject.settings.main". If this isn't          │
│                            provided, the DJANGO_SETTINGS_MODULE environment  │
│                            variable will be used.                            │
│ --pythonpath         PATH  A directory to add to the Python path, e.g.       │
│                            "/home/djangoprojects/myproject".                 │
│ --traceback                Raise on CommandError exceptions                  │
│ --show-locals              Print local variables in tracebacks.              │
│ --no-color                 Don't colorize the command output.                │
│ --force-color              Force colorization of the command output.         │
│ --skip-checks              Skip system checks.                               │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ create   Create an object.                                                   │
│ delete   Delete an object.                                                   │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

multi_create_help = """
 Usage: ./manage.py multi create [OPTIONS] NAME                                 
                                                                                
 Create an object.                                                              
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    name      TEXT  The name of the object to create. [required]            │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

multi_delete_help = """
 Usage: ./manage.py multi delete [OPTIONS] ID                                   
                                                                                
 Delete an object.                                                              
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    id      INTEGER  The id of the object to delete. [required].            │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

hierarchy_help = """
 Usage: ./manage.py hierarchy [OPTIONS] COMMAND [ARGS]...                       
                                                                                
 A more complex command that defines a hierarchy of subcommands.                
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --version                  Show program's version number and exit.           │
│ --settings           TEXT  The Python path to a settings module, e.g.        │
│                            "myproject.settings.main". If this isn't          │
│                            provided, the DJANGO_SETTINGS_MODULE environment  │
│                            variable will be used.                            │
│ --pythonpath         PATH  A directory to add to the Python path, e.g.       │
│                            "/home/djangoprojects/myproject".                 │
│ --traceback                Raise on CommandError exceptions                  │
│ --show-locals              Print local variables in tracebacks.              │
│ --no-color                 Don't colorize the command output.                │
│ --force-color              Force colorization of the command output.         │
│ --skip-checks              Skip system checks.                               │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ math   Do some math at the given precision.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

hierarchy_help_no_rich = """
Usage: ./manage.py hierarchy [OPTIONS] COMMAND [ARGS]...

  A more complex command that defines a hierarchy of subcommands.

Options:
  --version          Show program's version number and exit.
  --settings TEXT    The Python path to a settings module, e.g.
                     "myproject.settings.main". If this isn't provided, the
                     DJANGO_SETTINGS_MODULE environment variable will be used.
  --pythonpath PATH  A directory to add to the Python path, e.g.
                     "/home/djangoprojects/myproject".
  --traceback        Raise on CommandError exceptions
  --no-color         Don't colorize the command output.
  --force-color      Force colorization of the command output.
  --skip-checks      Skip system checks.
  --help             Show this message and exit.

Commands:
  math  Do some math at the given precision.
"""

hierarchy_math_help = """
 Usage: ./manage.py hierarchy math [OPTIONS] COMMAND [ARGS]...                  
                                                                                
 Do some math at the given precision.                                           
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --precision        INTEGER  The number of decimal places to output.          │
│                             [default: 2]                                     │
│ --help                      Show this message and exit.                      │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ divide     Divide the given numbers.                                         │
│ multiply   Multiply the given numbers.                                       │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

hierarchy_math_divide_help = """
 Usage: ./manage.py hierarchy math divide [OPTIONS] NUMERATOR DENOMINATOR       
                                                                                
 Divide the given numbers.                                                      
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    numerator        FLOAT  The numerator [required]                        │
│ *    denominator      FLOAT  The denominator [required]                      │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --floor    --no-floor      Use floor division [default: no-floor]            │
│ --help                     Show this message and exit.                       │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

hierarchy_math_multiply_help = """
 Usage: ./manage.py hierarchy math multiply [OPTIONS] NUMBERS...                
                                                                                
 Multiply the given numbers.                                                    
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    numbers      NUMBERS...  The numbers to multiply [required]             │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

hierarchy_math_help_no_rich = """
Usage: ./manage.py hierarchy math [OPTIONS] COMMAND [ARGS]...

  Do some math at the given precision.

Options:
  --precision INTEGER  The number of decimal places to output.  [default: 2]
  --help               Show this message and exit.

Commands:
  divide    Divide the given numbers.
  multiply  Multiply the given numbers.
"""

hierarchy_math_divide_help_no_rich = """
Usage: ./manage.py hierarchy math divide [OPTIONS] NUMERATOR DENOMINATOR

  Divide the given numbers.

Arguments:
  NUMERATOR    The numerator  [required]
  DENOMINATOR  The denominator  [required]

Options:
  --floor / --no-floor  Use floor division  [default: no-floor]
  --help                Show this message and exit.
"""

hierarchy_math_multiply_help_no_rich = """
Usage: ./manage.py hierarchy math multiply [OPTIONS] NUMBERS...

  Multiply the given numbers.

Arguments:
  NUMBERS...  The numbers to multiply  [required]

Options:
  --help  Show this message and exit.
"""


class ExampleTests(TestCase):
    settings = "tests.settings.examples"

    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_basic(self):
        observed_help = run_command(
            "basic", "--settings", self.settings, "--no-color", "--help"
        )[0].strip()
        self.assertGreater(
            similarity(
                observed_help, basic_help if rich_installed else basic_help_no_rich
            ),
            0.99,
        )

    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_multi(self):
        observed_help = run_command(
            "multi", "--settings", self.settings, "--no-color", "--help"
        )[0].strip()
        self.assertGreater(
            similarity(
                observed_help, multi_help if rich_installed else multi_help_no_rich
            ),
            0.99,
        )

        observed_help = run_command(
            "multi", "--settings", self.settings, "--no-color", "create", "--help"
        )[0].strip()
        self.assertGreater(
            similarity(
                observed_help,
                multi_create_help if rich_installed else multi_create_help_no_rich,
            ),
            0.99,
        )

        observed_help = run_command(
            "multi", "--settings", self.settings, "--no-color", "delete", "--help"
        )[0].strip()
        self.assertGreater(
            similarity(
                observed_help,
                multi_delete_help if rich_installed else multi_delete_help_no_rich,
            ),
            0.99,
        )

    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_hierarchy(self):
        observed_help = run_command(
            "hierarchy", "--settings", self.settings, "--no-color", "--help"
        )[0].strip()
        self.assertGreater(
            similarity(
                observed_help,
                hierarchy_help if rich_installed else hierarchy_help_no_rich,
            ),
            0.99,
        )

        observed_help = run_command(
            "hierarchy", "--settings", self.settings, "--no-color", "math", "--help"
        )[0].strip()
        self.assertGreater(
            similarity(
                observed_help,
                hierarchy_math_help if rich_installed else hierarchy_math_help_no_rich,
            ),
            0.99,
        )

        observed_help = run_command(
            "hierarchy",
            "--settings",
            self.settings,
            "--no-color",
            "math",
            "divide",
            "--help",
        )[0].strip()
        self.assertGreater(
            similarity(
                observed_help,
                hierarchy_math_divide_help
                if rich_installed
                else hierarchy_math_divide_help_no_rich,
            ),
            0.99,
        )

        observed_help = run_command(
            "hierarchy",
            "--settings",
            self.settings,
            "--no-color",
            "math",
            "multiply",
            "--help",
        )[0].strip()
        try:
            self.assertGreater(
                similarity(
                    observed_help,
                    hierarchy_math_multiply_help
                    if rich_installed
                    else hierarchy_math_multiply_help_no_rich,
                ),
                0.99,
            )
        except AssertionError:
            import ipdb

            ipdb.set_trace()

        self.assertEqual(
            run_command(
                "hierarchy",
                "--settings",
                self.settings,
                "--no-color",
                "math",
                "--precision",
                "4",
                "multiply",
                "3",
                "7",
                parse_json=False,
            )[0].strip(),
            "21.0000",
        )

        self.assertEqual(
            run_command(
                "hierarchy",
                "--settings",
                self.settings,
                "--no-color",
                "math",
                "--precision",
                "4",
                "divide",
                "3",
                "7",
                parse_json=False,
            )[0].strip(),
            "0.4286",
        )


class TyperExampleTests(ExampleTests):
    settings = "tests.settings.typer_examples"


@override_settings(
    INSTALLED_APPS=[
        "tests.apps.examples.completers",
        "tests.apps.examples.basic",
        "django_typer",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
    ]
)
class CompleterExampleTests(TestCase):
    app_labels_cmd = "app_labels"

    def test_app_labels_completer(self):
        from django_typer.management.commands.shellcompletion import (
            Command as ShellCompletion,
        )

        shellcompletion = get_command("shellcompletion", ShellCompletion).init(
            shell="zsh"
        )
        completions = shellcompletion.complete(f"{self.app_labels_cmd} ")
        self.assertTrue("contenttypes" in completions)
        self.assertTrue("completers" in completions)
        self.assertTrue("django_typer" in completions)
        self.assertTrue("admin" in completions)
        self.assertTrue("auth" in completions)
        self.assertTrue("sessions" in completions)
        self.assertTrue("messages" in completions)

        completions = shellcompletion.complete(f"{self.app_labels_cmd} a")

        self.assertTrue("admin" in completions)
        self.assertTrue("auth" in completions)

        stdout, stderr, retcode = run_command(
            self.app_labels_cmd,
            "--settings",
            "tests.settings.examples",
            "admin",
            "auth",
        )
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), str(["admin", "auth"]))


class CompleterTyperExampleTests(CompleterExampleTests):
    app_labels_cmd = "app_labels_typer"



import typing as t
import typer
from click import Context, Parameter
from click.shell_completion import CompletionItem
from django.apps import apps

from django_typer.management import Typer

app = Typer()


# the completer function signature must match this exactly
def complete_app_label(
    ctx: Context, param: Parameter, incomplete: str
) -> t.List[CompletionItem]:
    # don't offer apps that are already present as completion suggestions
    present = [app.label for app in (ctx.params.get(param.name or "") or [])]
    return [
        CompletionItem(app.label)
        for app in apps.get_app_configs()
        if app.label.startswith(incomplete) and app.label not in present
    ]


@app.command()
def handle(
    apps: t.Annotated[
        t.List[str],
        typer.Argument(
            help="The app label",
            # pass the completer function here
            shell_complete=complete_app_label,
        ),
    ],
):
    print(apps)



from django.core.management.base import BaseCommand, CommandParser
import sys
import json


class Command(BaseCommand):
    requires_system_checks = []
    requires_migrations_checks = False

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument("test_arg", type=int)
        parser.add_argument("--print", action="store_true")
        return super().add_arguments(parser)

    def handle(self, test_arg, print, **_):
        if print:
            return json.dumps(
                {
                    "no_typer": test_arg,
                    "print": print,
                    "modules": list(sys.modules.keys()),
                }
            )



from functools import partial

from django.core.management import get_commands

from . import these_strings

commands = partial(these_strings, lambda: get_commands().keys())
"""
A completer that completes management command names.

:param allow_duplicates: Whether or not to allow duplicate values. Defaults to False.
:return: A completer function.
"""



from django_typer.management import TyperCommand, command, group
from tests.apps.test_app.management.commands.order import Command as OrderCommand


class Command(OrderCommand):
    """
    Test that helps list commands in definition order.
    (This is different than click where the order is alphabetical by default)
    """

    @group()
    def bb(self):
        print("bb")

    @command()
    def aa(self):
        print("aa")

    @OrderCommand.d.group()
    def x(self):
        print("x")

    @command()
    def b(self):
        print("b")

    @OrderCommand.d.command()
    def i(self):
        print("i")

    @OrderCommand.d.command()
    def h(self):
        print("h")

    @command(help="Override handle")
    def handle(self):
        print("handle")

    @x.command()
    def z(self):
        print("z")

    @x.command()
    def y(self):
        print("y")



from django_typer.utils import called_from_module, called_from_command_definition

print(called_from_module())
print(called_from_module(look_back=0))
print(called_from_module(look_back=3))


def function():
    print(called_from_module())


def nested():
    def nested2():
        print(called_from_module())

    nested2()


class Command:
    called_from_defn = called_from_command_definition(look_back=0)


function()
nested()

print(Command.called_from_defn)



import sys
import typing as t
from pathlib import Path

import typer
import pluggy

from tests.apps.examples.plugins.media1.management.commands.backup import (
    Command as Backup,
)


class Command(Backup):  # inherit from the extended media backup command
    plugins = pluggy.PluginManager("backup")
    hookspec = pluggy.HookspecMarker("backup")
    hookimpl = pluggy.HookimplMarker("backup")

    # add a new command called files that delegates file backups to plugins
    @Backup.command()
    def files(self):
        """
        Backup app specific non-media files.
        """
        for archive in self.plugins.hook.backup_files(command=self):
            if archive:
                typer.echo(f"Backed up files to {archive}")


@Command.hookspec
def backup_files(command: Command) -> t.Optional[Path]:
    """
    A hook for backing up app specific files.

    Must return the path to the archive file or None if no files were backed up.

    :param command: the backup command instance
    :return: The path to the archived backup file
    """


Command.plugins.add_hookspecs(sys.modules[__name__])



import typing as t

from django.core.management.base import CommandError
from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import Typer
from tests.apps.examples.polls.models import Question as Poll

app = Typer(help="Closes the specified poll for voting")


@app.command()
def handle(
    self,
    poll_ids: t.Annotated[
        t.List[int],
        Argument(help=_("The database IDs of the poll(s) to close.")),
    ],
    delete: t.Annotated[
        bool, Option(help=_("Delete poll instead of closing it."))
    ] = False,
):
    for poll_id in poll_ids:
        try:
            poll = Poll.objects.get(pk=poll_id)
        except Poll.DoesNotExist:
            raise CommandError(f'Poll "{poll_id}" does not exist')

        poll.opened = False
        poll.save()

        self.stdout.write(
            self.style.SUCCESS(f'Successfully closed poll "{poll.id}"')
        )

        if delete:
            poll.delete()



from django.core.management import CommandError, call_command
from django.test import TestCase

from tests.utils import run_command


class UsageErrorTests(TestCase):
    def test_missing_parameter(self):
        result = run_command("missing")
        self.assertTrue("Test missing parameter." in result[0])
        self.assertTrue("arg1 must be given." in result[1])

        result = run_command("error")
        self.assertTrue("Test usage error behavior." in result[0])
        self.assertTrue("Missing parameter: arg1" in result[1])

        with self.assertRaises(CommandError):
            call_command("missing")

        with self.assertRaises(CommandError):
            call_command("error")

    def test_bad_param(self):
        result = run_command("error", "a")
        self.assertTrue("Test usage error behavior." in result[0])
        self.assertTrue("'a' is not a valid integer." in result[1])

        with self.assertRaises(CommandError):
            call_command("error", "a")

    def test_no_option(self):
        result = run_command("error", "--flg1")
        self.assertTrue("Test usage error behavior." in result[0])
        self.assertTrue("No such option: --flg1" in result[1])

        with self.assertRaises(CommandError):
            call_command("error", "--flg1")

    def test_bad_option(self):
        result = run_command("error", "--opt1", "d")
        self.assertTrue("Test usage error behavior." in result[0])
        self.assertTrue("'d' is not a valid integer." in result[1])

        with self.assertRaises(CommandError):
            call_command("error", "--opt1", "d")



import contextlib
from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class TestChaining(TestCase):
    def test_command_chaining(self):
        result = run_command(
            "chain", "command1", "--option=one", "command2", "--option=two"
        )[0]
        self.assertEqual(
            result.splitlines(), "command1\ncommand2\n['one', 'two']\n".splitlines()
        )

        result = run_command(
            "chain", "command2", "--option=two", "command1", "--option=one"
        )[0]
        self.assertEqual(
            result.splitlines(), "command2\ncommand1\n['two', 'one']\n".splitlines()
        )

        stdout = StringIO()
        with contextlib.redirect_stdout(stdout):
            result = call_command(
                "chain", "command2", "--option=two", "command1", "--option=one"
            )
        self.assertEqual(
            stdout.getvalue().splitlines(),
            "command2\ncommand1\n['two', 'one']\n".splitlines(),
        )
        self.assertEqual(result, ["two", "one"])

        chain = get_command("chain")
        self.assertEqual(chain.command1(option="one"), "one")
        self.assertEqual(chain.command2(option="two"), "two")



from django_typer.management import Typer

from .upstream_typer import app as upstream

# pass the upstream app as the first positional argument to Typer
app = Typer(upstream)


# override init
@app.callback()
def init():
    return "downstream:init"


# override sub1
@app.command()
def sub1():
    return "downstream:sub1"


# add a 3rd top level command
@app.command()
def sub3():
    return "downstream:sub3"


# add a new subcommand to grp1
@app.grp1.command()
def grp1_cmd2():
    return "downstream:grp1_cmd2"



import sys
import typing as t
from pathlib import Path

from tests.apps.examples.plugins.backup_files.management.commands.backup import (
    Command as Backup,
)


@Backup.hookimpl
def backup_files(command: Backup) -> t.Optional[Path]:
    # this is where you would put your custom file backup logic
    return command.output_directory / "files2.zip"


Backup.plugins.register(sys.modules[__name__])



from ..commands.upstream2_typer import app

# do not create a new app as with inheritance -
# instead work directly with the upstream app.


# override init
@app.callback()
def init():
    return "plugin:init"


# override sub1
@app.command()
def sub1():
    return "plugin:sub1"


# add a 3rd top level command
@app.command()
def sub3():
    return "plugin:sub3"


# add a new subcommand to grp1
@app.grp1.command()
def grp1_cmd2():
    return "plugin:grp1_cmd2"



import datetime
import shutil
import subprocess
import typing as t

import typer
from django.conf import settings

from tests.apps.examples.plugins.backup.management.commands import (
    backup_typer,
)


@backup_typer.app.command()
def environment(
    self,
    filename: t.Annotated[
        str,
        typer.Option(
            "-f",
            "--filename",
            help=("The name of the requirements file."),
        ),
    ] = "requirements.txt",
):
    """
    Capture the python environment using pip freeze.
    """

    output_file = self.output_directory / filename

    typer.echo(f"Capturing python environment to {output_file}")
    with output_file.open("w") as f:
        subprocess.run(["uv", "pip", "freeze"], stdout=f)


@backup_typer.app.command()
def database(self):
    """
    Backup the database by copying the sqlite file and tagging it with the
    current date.
    """
    db_file = self.output_directory / f"backup_{datetime.date.today()}.sqlite3"
    self.echo("Backing up database to {db_file}")
    shutil.copy(
        settings.DATABASES["default"]["NAME"],
        db_file,
    )



from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class TestReturnValues(TestCase):
    """
    Tests that overloaded group/command names work as expected.
    """

    def test_return_direct(self):
        return_cmd = get_command("return")
        self.assertEqual(return_cmd(), {"key": "value"})

    def test_return_cli(self):
        self.assertEqual(run_command("return")[0].strip(), str({"key": "value"}))

    def test_return_call(self):
        self.assertEqual(call_command("return"), {"key": "value"})



from django_typer.management import TyperCommand, COMMON_DEFAULTS
from typer.models import Context as TyperContext
from tests.utils import rich_installed


class Command(TyperCommand):
    help = "Test that django parameter suppression works as expected"

    suppressed_base_arguments = []

    def handle(self, ctx: TyperContext):
        assert self.__class__ is Command
        assert isinstance(ctx, TyperContext)
        assert not set(ctx.params.keys()).symmetric_difference(
            [
                key
                for key in COMMON_DEFAULTS.keys()
                if key
                not in ["hide_locals", *(["show_locals"] if not rich_installed else [])]
            ]
        )



from django_typer.management import Typer, TyperCommand

Command: TyperCommand

app = Typer()


@app.command()
def main(self, name: str):
    assert isinstance(self, TyperCommand)
    return {"name": name}



from django.core.management import CommandError

from django_typer.management import TyperCommand
from tests.utils import log_django_parameters


class Command(TyperCommand):
    help = "Test that django parameters work as expected"

    def handle(self, throw: bool = False):
        assert self.__class__ is Command
        log_django_parameters(self)
        if throw:
            raise CommandError("Test Exception")



from typing import Annotated

import typer
import json

from django_typer.management import TyperCommand
from django_typer.parsers.model import ReturnType
from django_typer.utils import duration_iso_string, model_parser_completer
from tests.apps.test_app.models import ShellCompleteTester
from django.db.models import QuerySet


class Command(TyperCommand, rich_markup_mode="rich"):
    def handle(
        self,
        query: Annotated[
            QuerySet[ShellCompleteTester],
            typer.Argument(
                **model_parser_completer(
                    ShellCompleteTester,
                    lookup_field="duration_field",
                    return_type=ReturnType.QUERY_SET,
                )
            ),
        ],
    ):
        assert self.__class__ is Command
        assert isinstance(query, QuerySet)
        return json.dumps(
            {obj.id: duration_iso_string(obj.duration_field) for obj in query}
        )



from tests.apps.test_app.management.commands.method_override import (
    Command as MethodOverride,
)


@MethodOverride.initialize(invoke_without_command=True)
def init(self):
    """
    Command1
    """
    assert self.__class__ is MethodOverride
    return "adapter0::init()"


@MethodOverride.command()
def cmd1(self):
    """
    Command1
    """
    assert self.__class__ is MethodOverride
    return "adapter0::cmd1()"


@MethodOverride.command()
def cmd2(self):
    """
    Command1
    """
    assert self.__class__ is MethodOverride
    return "adapter0::cmd2()"



"""
Test options metavar propagation and overrides
"""

from django_typer.management import TyperCommand, initialize


class Command(TyperCommand):
    @initialize()  # use default from TyperCommand
    def init(self):
        print("init")



from django.core.management import call_command
from django.test import TestCase

from django_typer.management import TyperCommand, get_command
from tests.utils import run_command


class TestHandle(TestCase):
    cmd = "handle"

    def test_handle_run(self):
        stdout, stderr, retcode = run_command(self.cmd)
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), self.cmd)

    def test_handle_call(self):
        self.assertEqual(call_command(self.cmd).strip(), self.cmd)

    def test_handle_direct(self):
        self.assertEqual(get_command(self.cmd, TyperCommand)(), self.cmd)


class TestHandleInheritOverrideAsCommand(TestHandle):
    cmd = "handle1"


class TestHandleAsCommand(TestHandle):
    cmd = "handle2"


class TestHandleOverride(TestHandle):
    cmd = "handle3"


class TestHandleAsCommandOverride(TestHandle):
    cmd = "handle4"


class TestNoHandle(TestCase):
    cmd = "handle_none"

    def test_no_handle_run(self):
        stdout, stderr, retcode = run_command(self.cmd)
        self.assertGreater(retcode, 0, msg=stderr)
        self.assertTrue("NotImplementedError" in stderr.strip())

    def test_no_handle_call(self):
        with self.assertRaises(NotImplementedError):
            call_command(self.cmd)

    def test_no_handle_direct(self):
        with self.assertRaises(NotImplementedError):
            get_command(self.cmd, TyperCommand)()



from django_typer.management import TyperCommand, group


class Command(TyperCommand):
    @group()
    def group1(self, common_option: bool = False):
        # you can define common options that will be available to all
        # subcommands of the group, and implement common initialization
        # logic here.
        ...

    @group()
    def group2(self): ...

    # attach subcommands to groups by using the command decorator on the group
    # function
    @group1.command()
    def grp1_subcommand1(self): ...

    @group1.command()
    def grp1_subcommand2(self): ...

    # groups can have subgroups!
    @group1.group()
    def subgroup1(self): ...

    @subgroup1.command()
    def subgrp_command(self): ...



from tests.apps.test_app.management.commands.adapted import (
    Command as Adapted,
)


@Adapted.command(name="echo")
def new_echo(self, msg1: str, msg2: str):
    """
    Echo both messages.
    """
    assert self.__class__ is Adapted
    self.echo(f"test_app2: {msg1} {msg2}")


@Adapted.command()
def no_self(int1: int, int2: int):
    assert isinstance(int1, int)
    assert isinstance(int2, int)
    print(f"test_app2: {int1, int2}")



# Generated by Django 5.1.7 on 2025-03-31 19:18

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ChoicesShellCompleteTester",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "char_choice",
                    models.CharField(
                        choices=[
                            ("git", "Git"),
                            ("svn", "Subversion (SVN)"),
                            ("hg", "Mercurial (Hg)"),
                            ("bzr", "Bazaar (Bzr)"),
                            ("cvs", "CVS"),
                            ("perforce", "Perforce"),
                            ("fossil", "Fossil"),
                            ("darcs", "Darcs"),
                            ("monotone", "Monotone"),
                        ],
                        db_index=True,
                        default=None,
                        max_length=12,
                        null=True,
                    ),
                ),
                (
                    "int_choice",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (1, "One"),
                            (2, "Two"),
                            (3, "Mercurial (Hg)"),
                            (4, "Bazaar (Bzr)"),
                            (5, "CVS"),
                            (6, "Perforce"),
                            (7, "Fossil"),
                            (8, "Darcs"),
                            (9, "Monotone"),
                            (10, "Ten"),
                            (11, "Eleven"),
                            (12, "Twelve"),
                        ],
                        db_index=True,
                        default=None,
                        null=True,
                    ),
                ),
                (
                    "ip_choice",
                    models.GenericIPAddressField(
                        choices=[
                            ("192.168.0.1", "Private Residential"),
                            ("10.0.0.1", "Private Gateway"),
                            ("172.16.0.1", "Corporate Network Gateway"),
                            ("8.8.8.8", "Google DNS"),
                            ("127.0.0.1", "Localhost IPv4"),
                            ("::1", "Localhost IPv6"),
                            ("2001:4860:4860::8888", "Google DNS IPv6"),
                            ("fe80::1", "Link-Local IPv6"),
                            ("fd00::1", "Private Network IPv6"),
                            ("2001:db8::1", "Documentation Example IPv6"),
                        ],
                        db_index=True,
                        default=None,
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ShellCompleteTester",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "char_field",
                    models.CharField(db_index=True, default="", max_length=15),
                ),
                ("text_field", models.TextField(default="")),
                (
                    "float_field",
                    models.FloatField(db_index=True, default=None, null=True),
                ),
                (
                    "decimal_field",
                    models.DecimalField(
                        db_index=True,
                        decimal_places=2,
                        default=None,
                        max_digits=10,
                        null=True,
                    ),
                ),
                ("uuid_field", models.UUIDField(default=None, null=True, unique=True)),
                ("binary_field", models.BinaryField(default=None, null=True)),
                (
                    "ip_field",
                    models.GenericIPAddressField(
                        db_index=True, default=None, null=True
                    ),
                ),
                (
                    "email_field",
                    models.EmailField(
                        db_index=True, default=None, max_length=254, null=True
                    ),
                ),
                ("url_field", models.URLField(db_index=True, default=None, null=True)),
                (
                    "file_field",
                    models.FileField(
                        db_index=True, default=None, null=True, upload_to="documents/"
                    ),
                ),
                (
                    "file_path_field",
                    models.FilePathField(db_index=True, default=None, null=True),
                ),
                (
                    "date_field",
                    models.DateField(db_index=True, default=None, null=True),
                ),
                (
                    "datetime_field",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                (
                    "time_field",
                    models.TimeField(db_index=True, default=None, null=True),
                ),
                (
                    "duration_field",
                    models.DurationField(db_index=True, default=None, null=True),
                ),
            ],
        ),
    ]



from django.utils.translation import gettext_lazy as _

from django_typer.management import TyperCommand, command


class Command(TyperCommand, help=_("2")):
    """
    5: Command class docstrings are the last resort for
    the upper level command help string.
    """

    help = _("3")

    # if an initializer is present it's help will be used for the command
    # level help

    @command(help=_("1"))
    def handle(self):
        """
        4: Function docstring is last priority and is not subject to
           translation.
        """



import re
import os
import pytest
from django.core.management import CommandError, call_command
from django.test import TestCase, override_settings
from django_typer.config import traceback_config, use_rich_tracebacks
from django_typer.utils import with_typehint

from tests.utils import rich_installed, run_command
import platform


class TracebackConfigTests(with_typehint(TestCase)):
    rich_installed = True

    def test_default_traceback(self):
        result = run_command("test_command1", "--no-color", "delete", "me", "--throw")[
            1
        ]
        self.assertIn("Traceback (most recent call last)", result)
        self.assertIn("Exception: This is a test exception", result)
        if self.rich_installed:
            self.assertIn("────────", result)
            # locals should not be present
            self.assertNotIn("name = 'me'", result)
            self.assertNotIn("throw = True", result)
            # by default we get only the last frame
            self.assertEqual(len(re.findall(r"\.py:\d+", result) or []), 1)

            hlp = run_command("test_command1", "--no-color", "--help")[0]
            # TODO - why are extra newlines inserted here on windows??
            if platform.system() == "Windows":
                hlp = hlp.replace("\r\n", "")
            self.assertNotIn("--hide-locals", hlp)
            self.assertIn("--show-locals", hlp)

            result = run_command(
                "test_command1",
                "--no-color",
                "--show-locals",
                "delete",
                "me",
                "--throw",
            )[1]
            self.assertIn("name = 'me'", result)
            self.assertIn("throw = True", result)
        else:
            self.assertNotIn("────────", result)

            hlp = run_command("test_command1", "--no-color", "--help")[0]
            self.assertNotIn("--hide-locals", hlp)
            self.assertNotIn("--show-locals", hlp)

    def test_tb_command_overrides(self):
        result = run_command(
            "test_tb_overrides", "--no-color", "delete", "me", "--throw"
        )[1]
        self.assertIn("Traceback (most recent call last)", result)
        self.assertIn("Exception: This is a test exception", result)
        if self.rich_installed:
            self.assertIn("────────", result)
            # locals should be present
            self.assertIn("name = 'me'", result)
            self.assertIn("throw = True", result)
            # should get a stack trace with files and line numbers
            self.assertGreater(len(re.findall(r"\.py:\d+", result) or []), 0)

            hlp = run_command("test_tb_overrides", "--no-color", "--help")[0]

            # TODO - why are extra newlines inserted here on windows??
            if platform.system() == "Windows":
                hlp = hlp.replace("\r\n", "")
            self.assertIn("--hide-locals", hlp)
            self.assertNotIn("--show-locals", hlp)

            result = run_command(
                "test_tb_overrides",
                "--no-color",
                "--hide-locals",
                "delete",
                "me",
                "--throw",
            )[1]
            self.assertNotIn("name = 'me'", result)
            self.assertNotIn("throw = True", result)
        else:
            self.assertNotIn("────────", result)

            hlp = run_command("test_tb_overrides", "--no-color", "--help")[0]
            self.assertNotIn("--hide-locals", hlp)
            self.assertNotIn("--show-locals", hlp)

    def test_turn_traceback_off_false(self):
        result = run_command(
            "test_command1",
            "--settings",
            "tests.settings.settings_tb_false",
            "delete",
            "me",
            "--throw",
        )[1]
        self.assertNotIn("────────", result)
        self.assertIn("Traceback (most recent call last)", result)
        self.assertIn("Exception: This is a test exception", result)

        hlp = run_command(
            "test_command1",
            "--settings",
            "tests.settings.settings_tb_false",
            "--no-color",
            "--help",
        )[0]
        self.assertFalse("--show-locals" in hlp)
        self.assertFalse("--hide-locals" in hlp)

    @override_settings(DT_RICH_TRACEBACK_CONFIG=False)
    def test_traceback_set_to_false(self):
        self.assertEqual(traceback_config(), {"show_locals": False})
        self.assertIs(use_rich_tracebacks(), False)

    @override_settings(DT_RICH_TRACEBACK_CONFIG=True)
    def test_traceback_set_to_true(self):
        self.assertEqual(traceback_config(), {"show_locals": False})
        self.assertIs(use_rich_tracebacks(), self.rich_installed)

    @override_settings(DT_RICH_TRACEBACK_CONFIG=None)
    def test_traceback_set_to_none(self):
        self.assertEqual(traceback_config(), {"show_locals": False})
        self.assertIs(use_rich_tracebacks(), False)

    def test_turn_traceback_off_none(self):
        result = run_command(
            "test_command1",
            "--settings",
            "tests.settings.settings_tb_none",
            "delete",
            "me",
            "--throw",
        )[1]
        self.assertNotIn("────────", result)
        self.assertIn("Traceback (most recent call last)", result)
        self.assertIn("Exception: This is a test exception", result)

        hlp = run_command(
            "test_command1",
            "--settings",
            "tests.settings.settings_tb_none",
            "--no-color",
            "--help",
        )[0]
        self.assertFalse("--show-locals" in hlp)
        self.assertFalse("--hide-locals" in hlp)

    def test_traceback_no_locals_short_false(self):
        result = run_command(
            "test_command1",
            "--no-color",
            "--settings",
            "tests.settings.settings_tb_change_defaults",
            "delete",
            "me",
            "--throw",
        )[1]
        self.assertIn("Traceback (most recent call last)", result)
        self.assertIn("Exception: This is a test exception", result)
        if self.rich_installed:
            self.assertIn("────────", result)
            self.assertGreater(len(re.findall(r"\.py:\d+", result) or []), 0)

            # locals should be present
            self.assertIn("name = 'me'", result)
            self.assertIn("throw = True", result)

            hlp = run_command(
                "test_command1",
                "--settings",
                "tests.settings.settings_tb_change_defaults",
                "--no-color",
                "--help",
            )[0]

            # TODO - why are extra newlines inserted here on windows??
            if platform.system() == "Windows":
                hlp = hlp.replace("\r\n", "")
            self.assertFalse("--show-locals" in hlp)
            self.assertTrue("--hide-locals" in hlp)
            result = run_command(
                "test_command1",
                "--no-color",
                "--hide-locals",
                "--settings",
                "tests.settings.settings_tb_change_defaults",
                "delete",
                "me",
                "--throw",
            )[1]

            self.assertNotIn("name = 'me'", result)
            self.assertNotIn("throw = True", result)
        else:
            self.assertNotIn("────────", result)

        self.assertNotIn("\x1b", result)

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="TODO --force-color not working on Windows",
    )
    def test_colored_traceback(self):
        result = run_command(
            "test_command1", "--force-color", "delete", "Brian", "--throw"
        )[1]
        if self.rich_installed:
            self.assertIn("\x1b", result)

        result = run_command(
            "test_command1", "--no-color", "delete", "Brian", "--throw"
        )[1]
        self.assertNotIn("\x1b", result)

        # the result of this depends on the terminal environment this runs in
        # result = run_command(
        #     "test_command1",
        #     "delete",
        #     "Brian",
        #     "--throw",
        # )[1]
        # self.assertIn("\x1b", result)


@pytest.mark.rich
@pytest.mark.skipif(not rich_installed, reason="rich not installed")
class TestTracebackConfig(TracebackConfigTests, TestCase):
    rich_installed = True

    @override_settings(DJ_RICH_TRACEBACK_CONFIG={"no_install": True})
    def test_tb_no_install(self):
        result = run_command(
            "test_command1",
            "--settings",
            "tests.settings.settings_tb_no_install",
            "delete",
            "me",
        )[1]
        self.assertIn("Traceback (most recent call last)", result)
        self.assertIn("Exception: Test ready exception", result)
        self.assertNotIn("────────", result)
        self.assertNotIn("── locals ──", result)

        hlp = run_command(
            "test_command1", "--settings", "tests.settings.settings_tb_no_install"
        )[1]
        self.assertFalse("--show-locals" in hlp)
        self.assertFalse("--hide-locals" in hlp)

    def test_rich_install(self):
        result = run_command(
            "test_command1",
            "--settings",
            "tests.settings.settings_throw_init_exception",
            "--no-color",
            "delete",
            "me",
        )[1]
        self.assertIn("Traceback (most recent call last)", result)
        self.assertIn("Exception: Test ready exception", result)
        self.assertIn("────────", result)
        self.assertNotIn("── locals ──", result)
        self.assertNotIn("\x1b", result)


@pytest.mark.no_rich
@pytest.mark.skipif(rich_installed, reason="rich installed")
class TestTracebackConfigNoRich(TracebackConfigTests, TestCase):
    rich_installed = False


class TestSettingsSystemCheck(TestCase):
    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_warning_thrown(self):
        result = run_command(
            "noop", "--settings", "tests.settings.settings_tb_bad_config"
        )[1]
        if rich_installed:
            self.assertIn(
                "tests.settings.settings_tb_bad_config: (django_typer.W001) DT_RICH_TRACEBACK_CONFIG",
                result,
            )
            self.assertIn(
                "HINT: Unexpected parameters encountered: unexpected_setting.", result
            )
        else:
            self.assertNotIn(
                "tests.settings.settings_tb_bad_config: (django_typer.W001) DT_RICH_TRACEBACK_CONFIG",
                result,
            )


class TracebackTests(TestCase):
    """
    Tests that show CommandErrors and UsageErrors do not result in tracebacks unless --traceback is set.

    Also make sure that sys.exit is not called when not run from the terminal
    (i.e. in get_command invocation or call_command).
    """

    def test_usage_error_no_tb(self):
        stdout, stderr, retcode = run_command("tb", "--no-color", "wrong")
        self.assertTrue("manage.py tb [OPTIONS] COMMAND [ARGS]" in stdout)
        self.assertTrue("No such command" in stderr)
        self.assertTrue(retcode > 0)

        stdout, stderr, retcode = run_command("tb", "--no-color", "error", "wrong")
        self.assertTrue("manage.py tb error [OPTIONS]" in stdout)
        self.assertTrue("Got unexpected extra argument" in stderr)
        self.assertTrue(retcode > 0)

        with self.assertRaises(CommandError):
            call_command("tb", "wrong")

        with self.assertRaises(CommandError):
            call_command("tb", "error", "wrong")

    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_usage_error_with_tb_if_requested(self):
        stdout, stderr, retcode = run_command(
            "tb", "--no-color", "--traceback", "wrong"
        )
        self.assertFalse(stdout.strip())
        self.assertTrue("Traceback" in stderr)
        if rich_installed:
            # self.assertTrue("───── locals ─────" in stderr)
            self.assertTrue("──────────── Traceback" in stderr)
        else:
            # self.assertFalse("───── locals ─────" in stderr)
            self.assertFalse("──────────── Traceback" in stderr)
        self.assertTrue("No such command 'wrong'" in stderr)
        self.assertTrue(retcode > 0)

        stdout, stderr, retcode = run_command(
            "tb", "--no-color", "--traceback", "error", "wrong"
        )
        self.assertFalse(stdout.strip())
        self.assertTrue("Traceback" in stderr)
        if rich_installed:
            # self.assertTrue("───── locals ─────" in stderr)
            self.assertTrue("──────────── Traceback" in stderr)
        else:
            # self.assertFalse("───── locals ─────" in stderr)
            self.assertFalse("──────────── Traceback" in stderr)
        self.assertFalse(stdout.strip())
        self.assertTrue("Got unexpected extra argument" in stderr)
        self.assertTrue(retcode > 0)

        with self.assertRaises(CommandError):
            call_command("tb", "--traceback", "wrong")

        with self.assertRaises(CommandError):
            call_command("tb", "--traceback", "error", "wrong")

    def test_click_exception_retcodes_honored(self):
        self.assertEqual(run_command("vanilla")[2], 0)
        self.assertEqual(run_command("vanilla", "--exit-code=2")[2], 2)

        self.assertEqual(run_command("tb", "exit")[2], 0)
        self.assertEqual(run_command("tb", "exit", "--code=2")[2], 2)

    def test_exit_on_call(self):
        with self.assertRaises(SystemExit):
            call_command("vanilla", "--help")

        with self.assertRaises(SystemExit):
            call_command("vanilla", "--exit-code", "0")

        with self.assertRaises(SystemExit):
            call_command("vanilla", "--exit-code", "1")

        with self.assertRaises(SystemExit):
            call_command("tb", "--help")

        with self.assertRaises(SystemExit):
            call_command("tb", "exit")

        with self.assertRaises(SystemExit):
            call_command("tb", "exit", "--code=1")



from django_typer.management import TyperCommand, command, initialize


class Command(TyperCommand):
    def finalize(self, **kwargs):
        assert isinstance(self, Command)
        return "finalized"

    @initialize(result_callback=finalize)
    def init(self):
        return "init"

    @command()
    def cmd(self):
        return self



from django_typer.management import group, initialize
from django_typer import types

from .interference import Command as Interference

from typing import Annotated

import typer


class Command(Interference):
    verbosity = 1

    flag1: bool

    @initialize()
    def init(
        self,
        verbosity: Annotated[
            int,
            typer.Option(
                help=(
                    "Verbosity level; 0=minimal output, 1=normal output, "
                    "2=verbose output, 3=very verbose output"
                ),
                show_choices=True,
                min=0,
                max=5,
                rich_help_panel=types.COMMON_PANEL,
            ),
        ] = verbosity,
    ):
        assert isinstance(self, Command)
        self.verbosity = verbosity

    @group()
    def grp2(self, flag1: bool = False):
        """Group 2, take a flag"""
        assert isinstance(self, Command)
        self.flag1 = flag1

    @grp2.command()
    def grp2_cmd1(self, cmd1_arg: str):
        assert isinstance(self, Command)
        return f"test_app::adapted2({self.verbosity})::grp2({self.flag1})::grp2_cmd1({cmd1_arg})"

    @grp2.command()
    def grp2_cmd2(self, cmd2_arg: str):
        assert isinstance(self, Command)
        return f"test_app::adapted2({self.verbosity})::grp2({self.flag1})::grp2_cmd2({cmd2_arg})"

    @Interference.grp1.command()
    def grp1_ext(self):
        """Inherit/extend grp1 with a cmd"""
        assert isinstance(self, Command)
        return f"test_app::adapted2({self.verbosity})::grp1({self.argg3})::grp1_ext()"



import shutil
import sys
from pathlib import Path
import platform
import typing as t

import pytest
from django.core.management import CommandError
from django.test import TestCase
from django_typer.shells.powershell import (
    PowerShellComplete,
    PwshComplete,
)

from tests.shellcompletion import (
    _ScriptCompleteTestCase,
    _InstalledScriptCompleteTestCase,
)


class _PowerShellMixin:
    shell = "powershell"
    profile: Path
    tabs = "\t"
    completer_class = PowerShellComplete

    environment = [
        f"[Console]::OutputEncoding = [System.Text.Encoding]::UTF8",
        f"$env:DJANGO_SETTINGS_MODULE='tests.settings.completion'",
        f'$env:PYTHONPATH = "$PWD;$env:PYTHONPATH"',
        f"{Path(sys.executable).absolute().parent / 'activate.ps1'}",
    ]

    @classmethod
    def setUpClass(cls) -> None:
        cls.profile = cls.completer_class().get_user_profile()
        return super().setUpClass()

    def verify_install(self, script=None, directory: t.Optional[Path] = None):
        if not script:
            script = self.manage_script
        self.assertTrue(self.profile.exists())
        self.assertTrue(
            f"Register-ArgumentCompleter -Native -CommandName {script} -ScriptBlock $scriptblock"
            in self.profile.read_text()
        )

    def verify_remove(self, script=None, directory: t.Optional[Path] = None):
        if not script:
            script = self.manage_script
        if self.profile.exists():
            contents = self.profile.read_text()
            self.assertFalse(
                f"Register-ArgumentCompleter -Native -CommandName {script} -ScriptBlock $scriptblock"
                in contents
            )
            self.assertTrue(contents)  # should have been deleted if it were empty

    @pytest.mark.skipif(platform.system() != "Windows", reason="Windows only test")
    def test_mixed_path_dividers(self):
        self.install()
        self.verify_install()
        completions = self.get_completions(
            "completion", "--path", "./src/django_typer\\comp"
        )
        self.assertIn(
            "src\\django_typer\\completers",
            completions,
        )
        self.remove()
        self.verify_remove()


@pytest.mark.skipif(
    shutil.which("powershell") is None, reason="powershell not available"
)
class PowerShellTests(_PowerShellMixin, _ScriptCompleteTestCase, TestCase):
    def test_shell_complete(self):
        with self.assertRaises(CommandError):
            self.install()

    @pytest.mark.rich
    @pytest.mark.no_rich
    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_rich_output(self): ...

    @pytest.mark.rich
    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_no_rich_output(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_settings_pass_through(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_pythonpath_pass_through(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_fallback(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_reentrant_install_uninstall(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_path_completion(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_mixed_path_dividers(self): ...


@pytest.mark.skipif(
    shutil.which("powershell") is None, reason="Powershell not available"
)
class PowerShellExeTests(_PowerShellMixin, _InstalledScriptCompleteTestCase, TestCase):
    pass


@pytest.mark.skipif(shutil.which("pwsh") is None, reason="pwsh not available")
class PWSHTests(_PowerShellMixin, _ScriptCompleteTestCase, TestCase):
    shell = "pwsh"
    completer_class = PwshComplete

    def test_shell_complete(self):
        with self.assertRaises(CommandError):
            self.install()

    @pytest.mark.rich
    @pytest.mark.no_rich
    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_rich_output(self): ...

    @pytest.mark.rich
    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_no_rich_output(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_settings_pass_through(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_pythonpath_pass_through(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_fallback(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_reentrant_install_uninstall(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_path_completion(self): ...

    @pytest.mark.skip(reason="powershell does not support script installations")
    def test_mixed_path_dividers(self): ...


@pytest.mark.skipif(shutil.which("pwsh") is None, reason="pwsh not available")
class PWSHExeTests(_PowerShellMixin, _InstalledScriptCompleteTestCase, TestCase):
    shell = "pwsh"
    completer_class = PwshComplete

    @pytest.mark.skipif(platform.system() != "Windows", reason="Windows only test")
    def test_mixed_path_dividers(self):
        self.install()
        self.verify_install()
        self.assertIn(
            "src\\django_typer\\completers",
            self.get_completions("completion", "--path", "./src/django_typer\\comp"),
        )
        self.remove()
        self.verify_remove()



import sys
import typing as t
from pathlib import Path

import typer
import pluggy

from django_typer.management import Typer
from tests.apps.examples.plugins.media1.management.commands import backup_typer

app = Typer(backup_typer.app)

# pluggy artifacts can live at module scope
plugins = pluggy.PluginManager("backup")
hookspec = pluggy.HookspecMarker("backup")
hookimpl = pluggy.HookimplMarker("backup")


# add a new command called files that delegates file backups to plugins
@app.command()
def files(self):
    """
    Backup app specific non-media files.
    """
    for archive in plugins.hook.backup_files(command=self):
        if archive:
            typer.echo(f"Backed up files to {archive}")


@hookspec
def backup_files(command) -> t.Optional[Path]:
    """
    A hook for backing up app specific files.

    Must return the path to the archive file or None if no files were backed up.

    :param command: the backup command instance
    :return: The path to the archived backup file
    """


plugins.add_hookspecs(sys.modules[__name__])



from django_typer.management import Typer

# Use extensions to the typer interface to improve clarity

app = Typer(chain=True)


@app.finalize()
def to_csv(results):
    return ", ".join(results)


@app.group(chain=True)
def grp():
    pass


@grp.finalize()
def to_upper_csv(results):
    return ", ".join([result.upper() for result in results])


@app.command()
def cmd1():
    return "result1"


@app.command()
def cmd2():
    return "result2"


@grp.command()
def cmd3():
    return "result3"


@grp.command()
def cmd4():
    return "result4"



from django_typer.management import TyperCommand, finalize


class Command(TyperCommand):
    @finalize()
    @staticmethod
    def final(result, traceback: bool = False):
        assert result == "handle"
        return f"finalized: {result} | traceback={traceback}"

    def handle(self):
        return "handle"



from django.test import TestCase

from tests.utils import run_command


class TestInheritance(TestCase):
    def test_handle_override(self):
        stdout, _, retcode = run_command("help_precedence8", "1", "2")
        self.assertEqual(retcode, 0)
        self.assertEqual(
            stdout,
            {
                "arg1": "1",
                "arg2": "2",
                "arg3": 0.5,
                "arg4": 1,
                "class": "<class 'tests.apps.test_app.management.commands.help_precedence8.Command'>",
            },
        )
        stdout, _, retcode = run_command("help_precedence16", "1", "2")
        self.assertEqual(retcode, 0)
        self.assertEqual(
            stdout,
            {
                "arg1": "1",
                "arg2": "2",
                "arg3": 0.5,
                "arg4": 1,
                "class": "<class 'tests.apps.test_app.management.commands.help_precedence16.Command'>",
            },
        )


class TestDiamondInheritance(TestCase):
    """
    Multiple inheritance works!! Even diamond inheritance!
    """

    def test_diamond_inheritance_run(self):
        stdout, _, retcode = run_command("inheritance3")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance2_1::init()" in stdout)

        stdout, _, retcode = run_command("inheritance3", "a")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance2_1::a()" in stdout)

        stdout, _, retcode = run_command("inheritance2_2", "a")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance2_2::a()" in stdout)

        stdout, _, retcode = run_command("inheritance3", "b")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance3::b()" in stdout)

        stdout, _, retcode = run_command("inheritance3", "c")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance2_2::c()" in stdout)

        stdout, _, retcode = run_command("inheritance3", "g2", "g2a")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance3::g2::g2a()" in stdout)

        stdout, _, retcode = run_command("inheritance1", "g")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance1::g()" in stdout)

        stdout, _, retcode = run_command("inheritance3", "g")
        self.assertGreater(retcode, 0)

        stdout, _, retcode = run_command("inheritance3", "g", "ga")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance2_1::g::ga()" in stdout)

        stdout, _, retcode = run_command("inheritance3", "g", "gb")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance2_1::g::gb()" in stdout)

        stdout, _, retcode = run_command("inheritance3", "g", "gc")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance2_1::g::gc()" in stdout)

        stdout, _, retcode = run_command("inheritance1", "g", "gc")
        self.assertGreater(retcode, 0)

        stdout, _, retcode = run_command("inheritance1", "g", "ga")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance1::g::ga()" in stdout)

        stdout, _, retcode = run_command("inheritance1", "g", "gb")
        self.assertEqual(retcode, 0)
        self.assertTrue("inheritance1::g::gb()" in stdout)



from django_typer.management import TyperCommand, command, group, initialize
from .inheritance1 import Command as Inheritance1Command


class Command(Inheritance1Command):
    """
    Inheritance2_1
    """

    @initialize(invoke_without_command=True)
    def init(self):
        assert isinstance(self, Command)
        return "inheritance2_1::init()"

    @command()
    def a(self):
        assert isinstance(self, Command)
        return "inheritance2_1::a()"

    @command()
    def b(self):
        assert isinstance(self, Command)
        return "inheritance2_1::b()"

    @group()
    def g(self):
        assert isinstance(self, Command)
        return "inheritance2_1::g()"

    @g.command()
    def ga(self):
        assert isinstance(self, Command)
        return "inheritance2_1::g::ga()"

    @g.command()
    def gb(self):
        assert isinstance(self, Command)
        return "inheritance2_1::g::gb()"

    @g.command()
    def gc(self):
        assert isinstance(self, Command)
        return "inheritance2_1::g::gc()"



from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class EmptyTests(TestCase):
    command = "empty"

    def test_empty_class_run(self):
        _, stderr, code = run_command(self.command)
        self.assertGreater(code, 0)
        self.assertTrue(
            f"NotImplementedError: No commands or command groups were registered on {self.command}"
            in stderr
        )

    def test_empty_class_call(self):
        with self.assertRaises(NotImplementedError):
            call_command(self.command)

    def test_empty_class_direct(self):
        with self.assertRaises(NotImplementedError):
            get_command(self.command)


class EmptyTyperTests(EmptyTests):
    command = "empty2"



from django.db import models
import django


def get_ip_choices():
    return [
        ("192.168.0.1", "Private Residential"),
        ("10.0.0.1", "Private Gateway"),
        ("172.16.0.1", "Corporate Network Gateway"),
        ("8.8.8.8", "Google DNS"),
        ("127.0.0.1", "Localhost IPv4"),
        ("::1", "Localhost IPv6"),
        ("2001:4860:4860::8888", "Google DNS IPv6"),
        ("fe80::1", "Link-Local IPv6"),
        ("fd00::1", "Private Network IPv6"),
        ("2001:db8::1", "Documentation Example IPv6"),
    ]


class ShellCompleteTester(models.Model):
    char_field = models.CharField(max_length=15, db_index=True, default="")
    text_field = models.TextField(default="")
    float_field = models.FloatField(db_index=True, default=None, null=True)
    decimal_field = models.DecimalField(
        db_index=True, default=None, null=True, max_digits=10, decimal_places=2
    )
    uuid_field = models.UUIDField(null=True, default=None, unique=True)

    binary_field = models.BinaryField(null=True, default=None)

    ip_field = models.GenericIPAddressField(null=True, default=None, db_index=True)

    email_field = models.EmailField(null=True, default=None, db_index=True)

    url_field = models.URLField(null=True, default=None, db_index=True)

    file_field = models.FileField(
        null=True, default=None, upload_to="documents/", db_index=True
    )

    file_path_field = models.FilePathField(null=True, default=None, db_index=True)

    date_field = models.DateField(null=True, default=None, db_index=True)

    datetime_field = models.DateTimeField(null=True, default=None, db_index=True)

    time_field = models.TimeField(null=True, default=None, db_index=True)

    duration_field = models.DurationField(null=True, default=None, db_index=True)


class ChoicesShellCompleteTester(models.Model):
    CHAR_CHOICES = [
        ("git", "Git"),
        ("svn", "Subversion (SVN)"),
        ("hg", "Mercurial (Hg)"),
        ("bzr", "Bazaar (Bzr)"),
        ("cvs", "CVS"),
        ("perforce", "Perforce"),
        ("fossil", "Fossil"),
        ("darcs", "Darcs"),
        ("monotone", "Monotone"),
    ]

    char_choice = models.CharField(
        max_length=12,
        choices=CHAR_CHOICES,
        db_index=True,
        default=None,
        null=True,
    )

    INT_CHOICES = [
        (1, "One"),
        (2, "Two"),
        (3, "Mercurial (Hg)"),
        (4, "Bazaar (Bzr)"),
        (5, "CVS"),
        (6, "Perforce"),
        (7, "Fossil"),
        (8, "Darcs"),
        (9, "Monotone"),
        (10, "Ten"),
        (11, "Eleven"),
        (12, "Twelve"),
    ]

    int_choice = models.PositiveSmallIntegerField(
        choices=INT_CHOICES,
        db_index=True,
        default=None,
        null=True,
    )

    IP_CHOICES = [
        ("192.168.0.1", "Private Residential"),
        ("10.0.0.1", "Private Gateway"),
        ("172.16.0.1", "Corporate Network Gateway"),
        ("8.8.8.8", "Google DNS"),
        ("127.0.0.1", "Localhost IPv4"),
        ("::1", "Localhost IPv6"),
        ("2001:4860:4860::8888", "Google DNS IPv6"),
        ("fe80::1", "Link-Local IPv6"),
        ("fd00::1", "Private Network IPv6"),
        ("2001:db8::1", "Documentation Example IPv6"),
    ]

    ip_choice = models.GenericIPAddressField(
        choices=IP_CHOICES,
        db_index=True,
        default=None,
        null=True,
    )



from .handle import Command as Handle


class Command(Handle):
    help = "Test various forms of handle override."

    def handle(self) -> str:
        assert self.__class__ is Command
        return "handle3"



# Generated by Django 4.2.13 on 2024-06-05 23:09

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Question",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("question_text", models.CharField(max_length=200)),
                (
                    "pub_date",
                    models.DateTimeField(verbose_name="date published"),
                ),
                ("opened", models.BooleanField(default=True)),
            ],
            options={
                "verbose_name": "Poll",
                "verbose_name_plural": "Polls",
            },
        ),
        migrations.CreateModel(
            name="Choice",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("choice_text", models.CharField(max_length=200)),
                ("votes", models.IntegerField(default=0)),
                (
                    "question",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="tests_apps_examples_polls.question",
                    ),
                ),
            ],
        ),
    ]



from typing import Any, TextIO
from django_typer.management import (
    TyperCommand,
    COMMON_DEFAULTS,
    initialize,
    command,
    group,
)
from typer.models import Context as TyperContext
from tests.utils import rich_installed


class Command(TyperCommand):
    help = "Test that django parameter suppression works as expected"

    suppressed_base_arguments = {"verbosity", "skip_checks", "traceback"}

    tb: bool

    @initialize(invoke_without_command=True)
    def init(self, ctx: TyperContext, traceback: bool = True):
        self.tb = traceback
        return self.check_context(ctx, traceback)

    @command()
    def cmd(self, ctx: TyperContext):
        return self.check_context(ctx, self.tb)

    @group(invoke_without_command=True)
    def subgroup(self, ctx: TyperContext, skip_checks: bool = True):
        return self.check_context(ctx, self.tb, skip_checks=skip_checks)

    def check_context(self, ctx, traceback, skip_checks=None):
        assert self.suppressed_base_arguments
        assert self.__class__ is Command
        assert isinstance(ctx, TyperContext)
        set(COMMON_DEFAULTS.keys())
        assert ctx.params
        assert "verbosity" not in ctx.params
        assert (
            "skip_checks" not in ctx.params
            if skip_checks is None
            else "skip_checks" in ctx.params
        )
        assert "traceback" in ctx.params
        for param in [
            key
            for key in COMMON_DEFAULTS.keys()
            if key
            not in ["hide_locals", *(["show_locals"] if not rich_installed else [])]
        ]:
            if param not in self.suppressed_base_arguments:
                assert param in ctx.params
        return f"traceback={traceback}, skipchecks={skip_checks}"



import typing as t

from django.apps import AppConfig, apps
from django.core.management import CommandError


def app_config(label: t.Union[str, AppConfig]):
    """
    A parser for app :attr:`~django.apps.AppConfig.label`. If the label is already
    an :class:`~django.apps.AppConfig` instance, the instance is returned. The label
    will be tried first, if that fails the label will be treated as the app
    :attr:`~django.apps.AppConfig.name`.

    .. code-block:: python

        import typing as t
        import typer
        from django_typer.management import TyperCommand
        from django_typer.parsers import app_config

        class Command(TyperCommand):

            def handle(
                self,
                django_apps: t.Annotated[
                    t.List[AppConfig],
                    typer.Argument(
                        parser=app_config,
                        help=_("One or more application labels.")
                    )
                ]
            ):
                ...

    :param label: The label to map to an AppConfig instance.
    :raises CommandError: If no matching app can be found.
    """
    if isinstance(label, AppConfig):
        return label
    try:
        return apps.get_app_config(label)
    except LookupError as err:
        for cfg in apps.get_app_configs():
            if cfg.name == label:
                return cfg

        raise CommandError(f"{label} does not match any installed app label.") from err


app_config.__name__ = "APP_LABEL"



from django.apps import AppConfig

from django_typer.utils import register_command_plugins


class Adapter1Config(AppConfig):
    name = "tests.apps.adapter0"
    label = "adapter0"
    verbose_name = "Adapter 0"

    def ready(self):
        from .management import adapters

        register_command_plugins(adapters)



from django_typer.management import TyperCommand, command
from typing import List


def finalize(self: List["Command"], **_):
    assert isinstance(self[0], Command)
    return "finalize".format(self)


class Command(TyperCommand, result_callback=finalize, chain=True):
    @command()
    def cmd1(self):
        return self

    @command()
    def cmd2(self):
        return self



import json

from django_typer.management import TyperCommand
from django_typer.utils import rich_installed


class Command(TyperCommand, rich_markup_mode="rich" if rich_installed else None):
    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        assert self.__class__ is Command
        opts = {"arg1": arg1, "arg2": arg2, "arg3": arg3, "arg4": arg4}
        return json.dumps(opts)



from django.apps import AppConfig

from django_typer.utils import register_command_plugins


class Adapter2Config(AppConfig):
    name = "tests.apps.adapter2"
    label = "adapter2"
    verbose_name = "Adapter 2"

    def ready(self):
        from .management import adapters

        register_command_plugins(adapters, ["adapted2"])



from tests.apps.test_app.management.commands.finalize_subgroups_inherit import (
    Command as FinalizeSubgroupsInherit,
)


# test that we can override a finalize method from a parent class
@FinalizeSubgroupsInherit.grp1.finalize()
def grp1_collect(self, result, **kwargs):
    assert isinstance(self, FinalizeSubgroupsInherit)
    return f"grp1_collect: {self.grp1_final(result, **kwargs)}"


@FinalizeSubgroupsInherit.grp2.finalize()
def grp2_collect(result, **kwargs):
    return f"grp2_collect: {FinalizeSubgroupsInherit.grp2_final(result, **kwargs)}"


@FinalizeSubgroupsInherit.grp1.command()
def cmd5(self):
    return "cmd5"



"""
The shellcompletion command is a Django_ management command that installs and removes
shellcompletion scripts for supported shells (bash, fish, zsh, powershell). This
command is also the entry point for running the completion logic and can be used to
debug completer code.

.. typer:: django_typer.management.commands.shellcompletion.Command:typer_app
    :prog: django-admin shellcompletion
    :width: 80
    :convert-png: latex

:func:`~django_typer.management.commands.shellcompletion.Command.install` invokes
typer's shell completion installation logic, but does have to patch the installed
scripts. This is because there is only one installation for all Django_ management
commands, not each individual command. The completion logic here will failover to
Django_'s builtin autocomplete if the command in question is not a
:class:`~django_typer.management.TyperCommand`. To promote compatibility with other
management command libraries or custom completion logic, a fallback completion function
can also be specified.
"""

import contextlib
import io
import os
import platform
import re
import sys
import typing as t
from pathlib import Path
from types import ModuleType

from click import get_current_context
from click.core import ParameterSource
from click.shell_completion import (
    CompletionItem,
    split_arg_string,  # pyright: ignore[reportPrivateImportUsage]
)
from django.core.management import CommandError, ManagementUtility
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from shellingham import ShellDetectionFailure
from typer import Argument, Option
from typer.main import get_command as get_typer_command

from django_typer.completers import these_strings
from django_typer.completers.path import import_paths
from django_typer.management import TyperCommand, command, get_command, initialize
from django_typer.shells import _completers
from django_typer.types import COMMON_PANEL
from django_typer.utils import detect_shell, get_usage_script, get_win_shell

DETECTED_SHELL = None

try:
    DETECTED_SHELL = detect_shell()[0]
except (ShellDetectionFailure, RuntimeError):  # pragma: no cover
    pass


def django_autocomplete(args: t.List[str], incomplete: str) -> t.List[CompletionItem]:
    # spoof bash environment variables
    # the first one is lopped off, so we insert a placeholder 0
    args = ["0", *args]
    if args[-1] != incomplete:
        args.append(incomplete)
    else:  # pragma: no cover
        pass
    os.environ["COMP_WORDS"] = " ".join(args)
    os.environ["COMP_CWORD"] = str(args.index(incomplete))
    os.environ["DJANGO_AUTO_COMPLETE"] = "1"
    dj_manager = ManagementUtility(args)
    capture_completions = io.StringIO()
    with contextlib.redirect_stdout(capture_completions):
        try:
            dj_manager.autocomplete()
        except SystemExit:
            pass
    return [
        CompletionItem(item) for item in capture_completions.getvalue().split() if item
    ]


class Command(TyperCommand):
    """
    This command installs autocompletion for the current shell. This command uses the
    typer/click autocompletion scripts to generate the autocompletion items, but monkey
    patches the scripts to invoke our bundled shell complete script which fails over to
    the django autocomplete function when the command being completed is not a
    :class:`~django_typer.management.TyperCommand`. When the django autocomplete
    function is used we also wrap it so that it works for any supported click/typer
    shell, not just bash.

    We also provide a remove command to easily remove the installed script.

    Great pains are taken to use the upstream dependency's shell completion logic. This
    is so advances and additional shell support implemented upstream should just work.
    However, it would be possible to add support for new shells here using the pluggable
    logic that click provides. It is probably a better idea however to add support for
    new shells at the typer level.

    Shell autocompletion can be brittle with every shell having its own quirks and
    nuances. We make a good faith effort here to support all the shells that typer/click
    support, but there can easily be system specific configuration issues that prevent
    this from working. In those cases users should refer to the online documentation for
    their specific shell to troubleshoot.
    """

    help = t.cast(str, _("Install autocompletion for the current shell."))

    # disable the system checks - no reason to run these for this one-off command
    requires_system_checks = []
    requires_migrations_checks = False

    # remove unnecessary django command base parameters - these just clutter the help
    suppressed_base_arguments = {
        "version",
        "skip_checks",
        "verbosity",
    }

    _shell: t.Optional[str] = DETECTED_SHELL
    shell_module: ModuleType

    ANSI_ESCAPE_RE = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

    _fallback: t.Optional[t.Callable[[t.List[str], str], t.List[CompletionItem]]] = None
    _manage_script: t.Optional[t.Union[str, Path]] = None

    color_default: bool = True

    @property
    def fallback(
        self,
    ) -> t.Optional[t.Callable[[t.List[str], str], t.List[CompletionItem]]]:
        return self._fallback

    @fallback.setter
    def fallback(self, fb: t.Optional[str]):
        try:
            self._fallback = import_string(fb) if fb else django_autocomplete
        except ImportError as err:
            raise CommandError(
                f"Unable to import fallback completion function: {err}"
            ) from err

    @property
    def fallback_import(self) -> t.Optional[str]:
        return (
            f"{self.fallback.__module__}.{self.fallback.__name__}"
            if self.fallback
            else None
        )

    @property
    def manage_script(self) -> t.Union[str, Path]:
        """
        Returns the name of the manage command as a string if it is available as a
        command on the user path. If it is a script that is not available as a command
        on the path it will return an absolute Path object to the script.
        """
        # We do not want to hardcode this as 'manage.py' because users are free to
        # rename and implement their own manage scripts! The safest way to do this is
        # therefore to fetch the name of the script that invoked the shellcompletion
        # command and determine if that script is available on the user's path or not.

        # Most shell's completion infrastructure works best if the commands are
        # available on the path. However, it is common for Django development to be done
        # in a virtual environment with a manage.py script being invoked directly as a
        # script. Completion should work in this case as well, but it does complicate
        # the installation for some shell's so we must first figure out which mode we
        # are in.
        if not self._manage_script:
            self.manage_script = None  # type: ignore
        return self._manage_script  # type: ignore

    @manage_script.setter
    def manage_script(self, script: t.Optional[str]):
        self._manage_script = get_usage_script(script)
        if isinstance(self._manage_script, Path):
            self._manage_script = self._manage_script.absolute()

    @property
    def manage_script_name(self) -> str:
        """
        Get the name of the manage script as a command available from the shell's path.
        """
        if isinstance(self.manage_script, Path):
            return self.manage_script.name
        return self.manage_script

    @property
    def shell(self) -> str:
        """
        Get the active shell. If not explicitly set, it first tries to find the shell
        in the environment variable shell complete scripts set and failing that it will
        try to autodetect the shell.
        """
        assert self._shell
        return self._shell

    @shell.setter
    def shell(self, shell: t.Optional[str]):
        """Set the shell to install autocompletion for."""
        if shell:
            self._shell = shell
        if self._shell is None:
            raise CommandError(
                "Unable to detect shell. Please specify the shell to install or "
                "remove autocompletion for."
            )
        elif self._shell == "cmd" and platform.system() == "Windows":
            try:
                self._shell = get_win_shell()
            except ShellDetectionFailure:  # pragma: no cover
                pass

    @property
    def shell_class(self):
        try:
            return _completers[self.shell]
        except KeyError as err:
            raise CommandError(f"Unsupported shell: {self.shell}") from err

    @initialize()
    def init(
        self,
        shell: t.Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _("The shell to use."),
                ),
                metavar="SHELL",
                shell_complete=these_strings(_completers.keys()),
            ),
        ] = DETECTED_SHELL,
        no_color: t.Annotated[
            t.Optional[bool],
            Option(
                "--no-color",
                help=t.cast(
                    str,
                    _(
                        "Filter terminal formatting control sequences out of completion"
                        " text."
                    ),
                ),
                rich_help_panel=COMMON_PANEL,
            ),
        ] = None,
        force_color: t.Annotated[
            bool,
            Option(
                "--force-color",
                help=t.cast(
                    str,
                    _(
                        "Allow terminal formatting control sequences in completion "
                        "text."
                    ),
                ),
                rich_help_panel=COMMON_PANEL,
            ),
        ] = False,
    ) -> "Command":
        self.shell = shell  # type: ignore[assignment]
        assert self.shell
        self.force_color = force_color
        self.no_color = (not self.shell_class.color) if no_color is None else no_color
        if self.force_color:
            self.no_color = False
        ctx = get_current_context(silent=True)
        self.color_default = (
            (
                ctx.get_parameter_source("no_color") is ParameterSource.DEFAULT
                and ctx.get_parameter_source("force_color") is ParameterSource.DEFAULT
            )
            if ctx
            else False
        )
        return self

    @command(
        help=t.cast(str, _("Install autocompletion for the current or given shell."))
    )
    def install(
        self,
        manage_script: t.Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The name of the django manage script to install autocompletion"
                        " for if different than the script used to invoke this command."
                    ),
                )
            ),
        ] = None,
        fallback: t.Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The python import path to a fallback complete function to use "
                        "when the completion command is not a TyperCommand."
                    ),
                ),
                shell_complete=import_paths,
            ),
        ] = None,
        template: t.Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The name of the template to use for the shell completion "
                        "script."
                    ),
                )
                # todo - add template name completer - see django-render-static
            ),
        ] = None,
        prompt: t.Annotated[
            bool,
            Option(
                "--no-prompt",
                help=t.cast(
                    str,
                    _("Do not ask for conformation before editing dotfiles."),
                ),
            ),
        ] = True,
    ) -> t.Optional[t.List[Path]]:
        """
        Install autocompletion for the given shell. If the shell is not specified, it
        will try to detect the shell. If the shell is not detected, it will fail.

        We run the upstream typer installation routines, with some augmentation.

        .. typer:: django_typer.management.commands.shellcompletion.Command:typer_app:install
            :prog: django-admin shellcompletion install
            :width: 85
            :convert-png: latex

        Returns the list of edited and/or created paths or None if no edits were made.
        """
        self.fallback = fallback  # type: ignore[assignment]
        self.manage_script = manage_script  # type: ignore[assignment]
        if isinstance(self.manage_script, Path):
            if not self.shell_class.supports_scripts:
                raise CommandError(
                    f"Shell {self.shell} does not support autocompletion for scripts "
                    f"that are not installed on the path. You must create an entry "
                    f"point for {self.manage_script_name}. See "
                    f"https://setuptools.pypa.io/en/latest/userguide/entry_point.html."
                )
            else:
                self.stdout.write(
                    self.style.WARNING(
                        f"It is not recommended to install tab completion for a script "
                        f"that is not installed on the path. This can be brittle. You "
                        f"should create an entry point for {self.manage_script_name}. "
                        f"See https://setuptools.pypa.io/en/latest/userguide/entry_point.html."
                    )
                )

        install_paths = self.shell_class(
            prog_name=str(manage_script or self.manage_script_name),
            command=self,
            template=template,
            color=not self.no_color or self.force_color,
            color_default=self.color_default,
        ).install(prompt=prompt)
        if install_paths:
            self.stdout.write(
                self.style.SUCCESS(
                    _("Installed autocompletion for {shell}").format(shell=self.shell)
                )
            )
        else:
            self.stdout.write(
                self.style.ERROR(
                    t.cast(str, _("Aborted shell completion installation."))
                )
            )
        return install_paths or None

    @command(
        help=t.cast(str, _("Uninstall autocompletion for the current or given shell."))
    )
    def uninstall(
        self,
        manage_script: t.Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The name of the django manage script to remove shell "
                        "completion for if different than the script used to invoke "
                        "this command."
                    ),
                )
            ),
        ] = None,
    ):
        """
        Remove the autocompletion for the given shell. If the shell is not specified,
        it will try to detect the shell. If the shell is not detected, it will fail.

        Since the installation routine is upstream we first run install to determine
        where the completion script is installed and then we remove it.

        .. typer:: django_typer.management.commands.shellcompletion.Command:typer_app:uninstall
            :prog: django-admin shellcompletion uninstall
            :width: 80
            :convert-png: latex

        """
        self.manage_script = manage_script  # type: ignore[assignment]
        self.shell_class(
            prog_name=str(manage_script or self.manage_script_name),
            command=self,
            color=not self.no_color or self.force_color,
            color_default=self.color_default,
        ).uninstall()
        self.stdout.write(
            self.style.WARNING(f"Uninstalled autocompletion for {self.shell}.")
        )

    @command(
        help=t.cast(str, _("Generate autocompletion for command string.")), hidden=False
    )
    def complete(
        self,
        command: t.Annotated[
            str,
            Argument(
                help=t.cast(
                    str, _("The command string to generate completion suggestions for.")
                ),
            ),
        ] = "",
        cursor: t.Annotated[
            t.Optional[int], Argument(help=t.cast(str, _("The cursor position.")))
        ] = None,
        fallback: t.Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The python import path to a fallback complete function to use "
                        "when the completion command is not a TyperCommand. By default,"
                        " the builtin django autocomplete function is used."
                    ),
                ),
                shell_complete=import_paths,
            ),
        ] = None,
    ):
        """
        We implement the shell complete generation script as a Django command because
        the Django environment needs to be bootstrapped for it to work. This also allows
        us to test completion logic in a platform agnostic way.

        .. tip::

            This command is super useful for debugging shell_complete logic. For example
            to enter into the debugger, we could set a breakpoint in our
            ``shell_complete`` function for the option parameter and then run the
            following command:

            .. code-block:: bash

                $ ./manage.py shellcompletion complete "./manage.py your_command --option "

        .. typer:: django_typer.management.commands.shellcompletion.Command:typer_app:complete
            :prog: django-admin shellcompletion complete
            :width: 80
            :convert-png: latex
        """
        command = command[:cursor] if cursor is not None else command
        args = split_arg_string(command.replace("\\", "\\\\"))
        if args:
            try:
                # lop the manage script off the front if it's there
                if args[0] == self.manage_script_name:
                    args = args[1:]
                elif Path(args[0]).resolve() == Path(sys.argv[0]).resolve():
                    args = args[1:]
            except (TypeError, ValueError, OSError):  # pragma: no cover
                pass

        def get_completion() -> str:
            if args:
                # we first need to figure out which command is being invoked
                # but we cant be sure which index the command name is at so
                # we try to fetch each in order and assume the first arg
                # that resolves to a command is the command
                cmd = None
                cmd_idx = -1
                try:
                    while cmd is None:
                        cmd_idx += 1
                        try:
                            cmd = get_command(args[cmd_idx])
                        except CommandError:
                            pass
                except IndexError:
                    if command.endswith(" "):
                        # unrecognized command
                        return ""
                    # fall through to fallback

                if isinstance(cmd, TyperCommand):
                    # this will exit out so no return is needed here
                    return self.shell_class(
                        cli=get_typer_command(cmd.typer_app),
                        ctx_args={"django_command": cmd},
                        prog_name="",
                        complete_var="",
                        command=self,
                        command_str=command,
                        command_args=args,
                        color=not self.no_color or self.force_color,
                        color_default=self.color_default,
                    ).complete()

            # only try to set the fallback if we have to use it
            self.fallback = fallback  # type: ignore[assignment]
            return self.shell_class(
                command=self,
                command_str=command,
                command_args=args,
                color=not self.no_color or self.force_color,
                color_default=self.color_default,
            ).complete()

        def strip_color(text: str) -> str:
            """
            Strip ANSI color codes from a string.
            """
            if self.no_color:
                return self.ANSI_ESCAPE_RE.sub("", text)
            return text

        env = os.environ.copy()
        try:
            return strip_color(get_completion())
        finally:
            os.environ.clear()
            os.environ.update(env)



from django.apps import AppConfig

from django_typer.utils import register_command_plugins


class HowtoApp(AppConfig):
    name = "tests.apps.howto"
    label = name.replace(".", "_")

    def ready(self):
        from .management import plugins

        register_command_plugins(plugins)



import typing as t

from typer import Argument, Option

from django_typer.management import TyperCommand
from django_typer.utils import model_parser_completer
from profiling.polls.models import Question as Poll


class Command(TyperCommand):
    help = "Closes the specified poll for voting."

    def handle(
        self,
        polls: t.Annotated[
            t.List[Poll],
            Argument(
                **model_parser_completer(Poll, help_field="question_text"),
                help="The database IDs of the poll(s) to close.",
            ),
        ],
        delete: t.Annotated[
            bool,
            Option(help="Delete poll instead of closing it."),
        ] = False,
    ):
        for poll in polls:
            poll.opened = False
            poll.save()
            self.stdout.write(
                self.style.SUCCESS(f'Successfully closed poll "{poll.pk}"')
            )
            if delete:
                poll.delete()



import json


from .help_precedence8 import Command as BaseCommand


# inherits docstring from base
class Command(BaseCommand):
    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        assert self.__class__ is Command
        opts = {"arg1": arg1, "arg2": arg2, "arg3": arg3, "arg4": arg4}
        return json.dumps(opts)



import typing as t
from functools import reduce

from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import Typer

app = Typer(
    help=_("A more complex command that defines a hierarchy of subcommands.")
)


math_grp = Typer(help=_("Do some math at the given precision."))

app.add_typer(math_grp, name="math")


@math_grp.callback()
def math(
    self,
    precision: t.Annotated[
        int, Option(help=_("The number of decimal places to output."))
    ] = 2,
):
    self.precision = precision


@math_grp.command(help=_("Multiply the given numbers."))
def multiply(
    self,
    numbers: t.Annotated[
        t.List[float], Argument(help=_("The numbers to multiply"))
    ],
):
    return f"{reduce(lambda x, y: x * y, [1, *numbers]):.{self.precision}f}"


@math_grp.command()
def divide(
    self,
    numerator: t.Annotated[float, Argument(help=_("The numerator"))],
    denominator: t.Annotated[float, Argument(help=_("The denominator"))],
    floor: t.Annotated[bool, Option(help=_("Use floor division"))] = False,
):
    """
    Divide the given numbers.
    """
    if floor:
        return str(numerator // denominator)
    return f"{numerator / denominator:.{self.precision}f}"



"""
A collection of useful utilities.
"""

import inspect
import os
import sys
import typing as t
from datetime import timedelta
from functools import partial
from importlib.util import find_spec
from pathlib import Path
from threading import local
from types import MethodType, ModuleType

from django.db.models import Model
from django.db.models.query import QuerySet

from .completers.model import ModelObjectCompleter
from .config import traceback_config
from .parsers.model import ModelObjectParser, ReturnType

# DO NOT IMPORT ANYTHING FROM TYPER HERE - SEE patch.py

__all__ = [
    "detect_shell",
    "get_usage_script",
    "get_current_command",
    "with_typehint",
    "register_command_plugins",
    "called_from_module",
    "called_from_command_definition",
    "duration_iso_string",
    "parse_iso_duration",
    "model_parser_completer",
    "rich_installed",
]


rich_installed = find_spec("rich") is not None and os.environ.get(
    "TYPER_USE_RICH", "true"
).lower() not in {"0", "false"}


def detect_shell(max_depth: int = 10) -> t.Tuple[str, str]:
    """
    Detect the current shell.

    :raises ShellDetectionFailure: If the shell cannot be detected
    :return: A tuple of the shell name and the shell command
    """
    from shellingham import ShellDetectionFailure
    from shellingham import detect_shell as _detect_shell

    try:
        return _detect_shell(max_depth=max_depth)
    except ShellDetectionFailure:
        login_shell = os.environ.get("SHELL", "")
        if login_shell:
            return (os.path.basename(login_shell).lower(), login_shell)
        raise


def get_usage_script(script: t.Optional[str] = None) -> t.Union[Path, str]:
    """
    Return the script name if it is on the path or the absolute path to the script
    if it is not.

    :param script: The script name to check. If None the current script is used.
    :return: The script name or the relative path to the script from cwd.
    """
    import shutil

    cmd_pth = Path(script or sys.argv[0])
    on_path: t.Optional[t.Union[str, Path]] = shutil.which(cmd_pth.name)
    on_path = on_path and Path(on_path)
    if (
        on_path
        and on_path.is_absolute()
        and (on_path == cmd_pth.absolute() or not cmd_pth.is_file())
    ):
        return cmd_pth.name
    try:
        return cmd_pth.absolute().relative_to(Path(os.getcwd()))
    except ValueError:
        return cmd_pth.absolute()


_command_context = local()


def get_current_command() -> t.Optional["TyperCommand"]:  # type: ignore  # noqa: F821
    """
    Returns the current typer command. This can be used as a way to
    access the current command object from anywhere if we are executing
    inside of one from higher on the stack. We primarily need this because certain
    monkey patches are required in typer code - namely for enabling/disabling
    color based on configured parameters.

    This function is thread safe.

    This is analogous to click's get_current_context but for
    command execution.

    :return: The current typer command or None if there is no active command.
    """
    try:
        return t.cast("TyperCommand", _command_context.stack[-1])  # type: ignore  # noqa: F821
    except (AttributeError, IndexError):
        pass
    return None


T = t.TypeVar("T")  # pylint: disable=C0103


def with_typehint(baseclass: t.Type[T]) -> t.Type[T]:
    """
    Type hinting mixin inheritance is really annoying. The current
    canonical way is to use Protocols but this is prohibitive when
    the super classes already exist and are extensive. All we're
    trying to do is let our type checker know about super() methods
    etc - this is a simple way to do that.
    """
    if t.TYPE_CHECKING:
        return baseclass  # pragma: no cover
    return object  # type: ignore


_command_plugins: t.Dict[str, t.List[ModuleType]] = {}


def register_command_plugins(
    package: ModuleType, commands: t.Optional[t.List[str]] = None
):
    """
    Register a command plugin for the given command within the given package.

    For example, use this in your AppConfig's ready() method:

    .. code-block:: python

        from django.apps import AppConfig
        from django_typer.utils import register_command_plugins


        class MyAppConfig(AppConfig):
            name = "myapp"

            def ready(self):
                from .management import plugins

                register_command_plugins(plugins)


    :param package: The package the command extension module resides in
    :param commands: The names of the commands/modules, if not provided, all modules
        in the package will be registered as plugins
    """
    import pkgutil

    commands = commands or [
        module[1].split(".")[-1]
        for module in pkgutil.iter_modules(package.__path__, f"{package.__name__}.")
    ]
    for command in commands:
        _command_plugins.setdefault(command, [])
        if package not in _command_plugins[command]:
            _command_plugins[command].append(package)


def _load_command_plugins(command: str) -> int:
    """
    Load any plugins for the given command by loading the registered
    modules in registration order.

    :param command: The name of the command
    :return: The number of plugins loaded.
    """
    plugins = _command_plugins.get(command, [])
    if plugins:
        import importlib

        for ext_pkg in reversed(plugins):
            try:
                importlib.import_module(f"{ext_pkg.__name__}.{command}")
            except (ImportError, ModuleNotFoundError) as err:
                raise ValueError(
                    f"No extension module was found for command {command} in "
                    f"{ext_pkg.__path__}."
                ) from err
        # we only want to do this once
        del _command_plugins[command]
    return len(plugins)


def _check_call_frame(frame_name: str, look_back=1) -> bool:
    """
    Returns True if the stack frame one frame above where this function has the given
    name.

    :param frame_name: The name of the frame to check for
    """
    frame = inspect.currentframe()
    for _ in range(0, look_back + 1):
        if not frame:
            break
        frame = frame.f_back
    if frame:
        return frame.f_code.co_name == frame_name
    return False


called_from_module = partial(_check_call_frame, "<module>")
called_from_command_definition = partial(_check_call_frame, "Command")


def is_method(
    func_or_params: t.Optional[t.Union[t.Callable[..., t.Any], t.List[str]]],
) -> t.Optional[bool]:
    """
    This logic is used to to determine if a function should be bound as a method
    or not. Right now django-typer will treat module scope functions as methods
    when binding to command classes if they have a first argument named 'self'.

    :param func: The function to check or a list of parameter names, or None
    :return: True if the function should be considered a method, False if not and None
        if undetermined.
    """
    # ##############
    # Remove when python 3.9 support is dropped
    func_or_params = getattr(func_or_params, "__func__", func_or_params)
    ##############
    if func_or_params:
        params = (
            list(inspect.signature(func_or_params).parameters)
            if callable(func_or_params)
            else func_or_params
        )
        if params:
            return params[0] == "self"
        return isinstance(func_or_params, MethodType)
    return None


def accepts_var_kwargs(func: t.Callable[..., t.Any]) -> bool:
    """
    Determines if the given function accepts variable keyword arguments.
    """
    for param in reversed(list(inspect.signature(func).parameters.values())):
        return param.kind is inspect.Parameter.VAR_KEYWORD
    return False


def accepted_kwargs(
    func: t.Callable[..., t.Any], kwargs: t.Dict[str, t.Any]
) -> t.Dict[str, t.Any]:
    """
    Return the named keyword arguments that are accepted by the given function.
    """
    if accepts_var_kwargs(func):
        return kwargs
    param_names = set(inspect.signature(func).parameters.keys())
    return {k: v for k, v in kwargs.items() if k in param_names}


def get_win_shell() -> str:
    """
    The way installed python scripts are wrapped on Windows means shellingham will
    detect cmd.exe as the shell. This function will attempt to detect the correct shell,
    usually either powershell (<=v5) or pwsh (>=v6).

    :raises ShellDetectionFailure: If the shell cannot be detected
    :return: The name of the shell, either 'powershell' or 'pwsh'
    """
    import json
    import shutil
    import subprocess

    from shellingham import ShellDetectionFailure

    pwsh = shutil.which("pwsh") or shutil.which("powershell")
    if pwsh:
        try:
            ps_command = """
            $parent = Get-CimInstance -Query "SELECT * FROM Win32_Process WHERE ProcessId = {pid}";
            $parentPid = $parent.ParentProcessId;
            $parentInfo = Get-CimInstance -Query "SELECT * FROM Win32_Process WHERE ProcessId = $parentPid";
            $parentInfo | Select-Object Name, ProcessId | ConvertTo-Json -Depth 1
            """
            pid = os.getpid()
            while True:
                result = subprocess.run(
                    [pwsh, "-NoProfile", "-Command", ps_command.format(pid=pid)],
                    capture_output=True,
                    text=True,
                ).stdout.strip()
                if not result:
                    break
                process = json.loads(result)
                if "pwsh" in process.get("Name", ""):
                    return "pwsh"
                elif "powershell" in process.get("Name", ""):
                    return "powershell"
                pid = process["ProcessId"]

        except Exception as e:  # pragma: no cover
            raise ShellDetectionFailure("Unable to detect windows shell") from e

    raise ShellDetectionFailure("Unable to detect windows shell")


def parse_iso_duration(duration: str) -> t.Tuple[timedelta, t.Optional[str]]:
    """
    Progressively parse an ISO8601 duration type - can be a partial
    duration string. If it is a partial duration string with an ambiguous
    trailing number, the number will be returned as the second value of the
    tuple.

    .. note::
        We use a subset of ISO8601, the supported markers are D, H, M, S.

    :return: A tuple of the parsed duration and the ambiguous trailing number
    """
    import re

    original = duration
    duration = duration.upper()

    sign = -1 if duration.startswith("-") else 1
    duration = duration.lstrip("-").lstrip("+").lstrip("P")

    ambiguous: t.Optional[str] = None

    class Incomplete(Exception):
        value: str

        def __init__(self, value: str):
            self.value = value

    def eat(markers: t.Sequence[str], interpret=lambda x: (int(x), True)) -> int:
        nonlocal duration
        if duration:
            match = re.match(r"(\d+)(.)?", duration)
            if match and match.group(2) in markers:
                duration = duration[match.end() :]
                return interpret(match.group(1))[0]
            if match and not match.group(2):
                duration = duration[match.end() :]
                value, ambig = interpret(match.group(1))
                if not ambig:
                    return value
                raise Incomplete(match.group(1))
        return 0

    days = 0
    hours = 0
    minutes = 0
    seconds = 0
    microseconds = 0

    # examples: 1 T1
    try:
        days = eat(("D",))
    except Incomplete as incomplete:
        ambiguous = incomplete.value

    duration = duration.lstrip("T")

    try:
        hours = eat(("H",))
        minutes = eat(("M",))
        seconds = eat((".", "S"))
        microseconds = eat(("S",), lambda x: (int(f"{x:0<6}"), len(x) < 6))
    except Incomplete as incomplete:
        ambiguous = incomplete.value

    if duration:
        # if the string was a valid full or partial duration all characters
        # should have been consumed
        raise ValueError(f"Invalid ISO 8601 duration format: {original}")

    return sign * timedelta(
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        microseconds=microseconds,
    ), ambiguous


def duration_iso_string(duration: timedelta) -> str:
    """
    Return an ISO8601 duration string from a timedelta. This differs from
    the Django implementation in that zeros are elided.
    """
    if not duration:
        return "PT0S"
    sign = "-" if duration < timedelta() else ""
    if sign:
        duration *= -1
    days = duration.days
    hours, seconds = divmod(duration.seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    time_parts = []
    day_str = ""
    if days:
        day_str = f"{abs(days)}D"
    if hours:
        time_parts.append(f"{abs(hours)}H")
    if minutes:
        time_parts.append(f"{abs(minutes)}M")
    if seconds or duration.microseconds:
        if duration.microseconds:
            time_parts.append(f"{abs(seconds)}.{abs(duration.microseconds):0>6}S")
        else:
            time_parts.append(f"{abs(seconds)}S")
    time_str = ""
    if time_parts:
        time_str = "T" + "".join(time_parts)
    return f"{sign}P{day_str}{time_str}"


def model_parser_completer(
    model_or_qry: t.Union[t.Type[Model], QuerySet],
    lookup_field: t.Optional[str] = None,
    case_insensitive: bool = False,
    help_field: t.Optional[str] = ModelObjectCompleter.help_field,
    query: t.Optional[ModelObjectCompleter.QueryBuilder] = None,
    limit: t.Optional[int] = ModelObjectCompleter.limit,
    distinct: bool = ModelObjectCompleter.distinct,
    on_error: t.Optional[ModelObjectParser.error_handler] = ModelObjectParser.on_error,
    order_by: t.Optional[t.Union[str, t.Sequence[str]]] = None,
    return_type: ReturnType = ModelObjectParser.return_type,
) -> t.Dict[str, t.Any]:
    """
    A factory function that returns a dictionary that can be used to specify
    a parser and completer for a typer.Option or typer.Argument. This is a
    convenience function that can be used to specify the parser and completer
    for a model object in one go.

    .. code-block:: python

        def handle(
            self,
            obj: t.Annotated[
                ModelClass,
                typer.Argument(
                    **model_parser_completer(ModelClass, 'field_name'),
                    help=_("Fetch objects by their field_names.")
                ),
            ]
        ):
            ...


    :param model_or_qry: the model class or QuerySet to use for lookup
    :param lookup_field: the field to use for lookup, by default the primary key
    :param case_insensitive: whether to perform case insensitive lookups and
        completions, default: False
    :param help_field: the field to use for help output in completion suggestions,
        by default no help will be provided
    :param query: a callable that will be used to build the query for completions,
        by default the query will be reasonably determined by the field type
    :param limit: the maximum number of completions to return, default: 50
    :param distinct: whether to filter out already provided parameters in the
        completion suggestions, True by default
    :param on_error: a callable that will be called if the parser lookup fails
        to produce a matching object - by default a CommandError will be raised
    :param return_type: An enumeration switch to return either a model instance,
        queryset or model field value type.
    """
    return {
        "parser": ModelObjectParser(
            model_or_qry if inspect.isclass(model_or_qry) else model_or_qry.model,  # type: ignore
            lookup_field,
            case_insensitive=case_insensitive,
            on_error=on_error,
            return_type=return_type,
        ),
        "shell_complete": ModelObjectCompleter(
            model_or_qry,
            lookup_field,
            case_insensitive=case_insensitive,
            help_field=help_field,
            query=query,
            limit=limit,
            distinct=distinct,
            order_by=order_by,
        ),
    }


def install_traceback(tb_config: t.Optional[t.Dict[str, t.Any]] = None):
    from .config import use_rich_tracebacks

    if not use_rich_tracebacks():
        return

    import rich
    from rich import traceback
    from typer import main as typer_main

    tb_config = tb_config or traceback_config()

    # install rich tracebacks if we've been configured to do so (default)
    no_color = "NO_COLOR" in os.environ
    force_color = "FORCE_COLOR" in os.environ
    traceback.install(
        console=tb_config.pop(
            "console",
            (
                rich.console.Console(
                    stderr=True,
                    no_color=no_color,
                    force_terminal=(
                        False if no_color else force_color if force_color else None
                    ),
                )
                if no_color or force_color
                else None
            ),
        ),
        **{
            param: value
            for param, value in tb_config.items()
            if param in set(inspect.signature(traceback.install).parameters.keys())
        },
    )
    # typer installs its own exception hook and it falls back to the sys hook -
    # depending on when typer was imported it may have the original fallback system hook
    # or our installed rich one - we patch it here to make sure!
    typer_main._original_except_hook = sys.excepthook
